
__all__ = [
    'get_color_fader',
    'training_curves', # should be training_curves for future?
    'plot_quadrant' , # should be plot_quadrant for future?
    'plot_correction', 
]

import os
import glob
import json
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import rootplotlib as rpl
import array

from Gaugi import Logger, expand_folders
from Gaugi.macros import *
from copy import copy

from kolmov.utils.constants import str_etbins_zee, str_etabins
from ROOT import kBlackBody, TCanvas, TGraphErrors, TLine, gStyle, kBlue, kBlack



def get_color_fader( c1, c2, n ):
    def color_fader(c1,c2,mix=0): #fade (linear interpolate) from color c1 (at mix=0) to c2 (mix=1)
        c1=np.array(mpl.colors.to_rgb(c1))
        c2=np.array(mpl.colors.to_rgb(c2))
        return mpl.colors.to_hex((1-mix)*c1 + mix*c2)
    return [ color_fader(c1,c2, frac) for frac in np.linspace(0,1,n) ]




class training_curves( Logger ):


    def __init__(self, path, model_idx , str_et_bins = str_etbins_zee,
                                         str_eta_bins = str_etabins
                                         ):

        '''
        This class is a little monitoring tool for saphyra trained models.
        Its use the dumped json history of a trained model. If you don't have this json
        use the dump_all_traiplot_monitoringn_history from base_table in kolmov.

        Arguments:
        - path_to_history: the path to json history files

        Ex.: /my_volume/my_history_files

        - model_idx: the index of model that you want extract the monitoring info.
        - str_et_bins: a list which contains the et boundaries.
        The default values are the zee et binning, other are in constants.
        '''

        Logger.__init__( self )
        self.plot_names = {
                            'loss' : 'Loss Function Evolution',
                            'sp'   : r'$SP$ Index Evolution',
                            'pd'  : r'$P_D$ Evolution',
                            'fa'  : r'$F_A$ Evolution',
                            }

        self.h_dict         = self.load(path, model_idx)
        self.sort_name_dict =  {idx : 'Fold %i' %(idx+1) for idx in range(10)}
        self.et_range = str_et_bins
        self.eta_range = str_eta_bins



    def load(self, basepath, model_idx):

        '''
        This method will open all the histories that was grouped in the initialize method,
        and put then into a dictionary in order to make easier to manipulate. Usually used
        for histories dumped from best inists table
        '''
        paths = expandFolders( basepath )
        MSG_INFO( self, "Reading %d files...", len(paths) )
        h_dict = dict()
        for path in paths:
            if path.split('.')[-1] != 'json':
                continue
            
            with open(path) as f:
                obj = dict(eval(json.load(f)))
                key = 'et%d_eta%d_sort_%d' %(obj['loc']['et_bin'], obj['loc']['eta_bin'], obj['loc']['sort'])
                if obj['loc']['model_idx'] != model_idx:
                  continue
                h_dict[key] = obj
        return h_dict




    def plot_training_curves(self, et_bin, eta_bin, plot_path, plot_name):
        '''
        Arguments:
        - et_bin: the energy bin index.
        - eta_bin: the pseudorapidity bin index.
        - plot_path: the path to save the plot.
        - plot_name: the plot name.

        Ex.:
        # initialize the base_ploter class
        plot_tool = plot_monitoring('my_path/awesome_history', model_idx=0)
        # now plot some bin
        plot_tool.plot_monitoring_curves(et_bin=0,
                                         eta_bin=0,
                                         plot_path=my_save_path,
                                         plot_name=a_very_meaningful_name)
        '''
        if not os.path.exists( plot_path ):
          os.makedirs( plot_path )

        # a easy way to lock only in the necessaries keys
        wanted_h_keys = ['et%i_eta%i_sort_%i' %(et_bin, eta_bin, i) for i in range(10)]


        # train indicators (today we use only this)
        train_indicators = [('loss', 'val_loss'), ('sp', 'val_sp'), ('pd', 'val_pd'), ('fa', 'val_fa')]
        # create a subplots and fill then.
        fig, ax = plt.subplots( nrows=len(wanted_h_keys),
                                ncols=len(train_indicators),
                                figsize=(25,20))
        fig.subplots_adjust(top=0.95, hspace=0.7)
        # loop over the sorts
        for idx, isort in enumerate(wanted_h_keys):
            isort_k = np.int(isort.split('_')[-1])
            # loop over the train indicators
            for jdx, train_ind in enumerate(train_indicators):
                if isinstance(train_ind, tuple):
                    ax[idx, jdx].set_title('%s - %s' %(self.plot_names[train_ind[0]],
                                                       self.sort_name_dict[isort_k]))
                    ax[idx, jdx].set_xlabel('Epochs')
                    ax[idx, jdx].plot(self.h_dict[isort][train_ind[0]], c='r', label='Train Step')
                    ax[idx, jdx].plot(self.h_dict[isort][train_ind[1]], c='b', label='Validation Step')
                    ax[idx, jdx].axvline(x=len(self.h_dict[isort][train_ind[1]])-25, label='Best Epoch',
                                         c='black', lw=1.25, ls='--')
                    ax[idx, jdx].legend()
                    ax[idx, jdx].grid()
                else:
                    ax[idx, jdx].set_title('%s - %s' %(self.plot_names[train_ind],
                                                       self.sort_name_dict[isort_k]))
                    ax[idx, jdx].axvline(x=self.h_dict[isort]['max_sp_best_epoch_val'][-1], label='Best Epoch',
                                         c='black', lw=1.25, ls='--')
                    ax[idx, jdx].set_xlabel('Epochs')
                    ax[idx, jdx].plot(self.h_dict[isort][train_ind], c='g', label='Validation Step')
                    ax[idx, jdx].grid()
        # set the figure name
        fig.suptitle(r'Monitoring Train Plot - $E_T$: %s | $\eta$: %s'
                     %(self.et_range[et_bin], self.eta_range[eta_bin]), fontsize=20)
        # save figure
        fig.savefig(os.path.join(plot_path, '%s.png' %(plot_name)),
                    dpi=300,
                    bbox_inches='tight', facecolor='white')
        plt.close();
        return




#
# plot quadrant histograms
#
def plot_quadrant(df, plot_configs, output_path='/volume'):
    '''
    This function will provide a simple implementation of prometheus quadrant analysis
    applying conditions on Dataframe.

    Arguments:
    - df: the Dataframe used to extract information;
    - plot_configs: a dictionary which contains the all plot information and the
    condition to produce the plot;
    - outpu_path: the path to save the plot.
    '''
    for ivar in plot_configs.keys():
        print('Plotting %s... ' %(ivar))
        local_config = plot_configs[ivar]
        # loop over variables
        var_cond1 = df[((local_config['cond1']) &\
                        (local_config['common_cond']))][ivar].values
        var_cond1 = var_cond1[((var_cond1 >= local_config['low_edge']) &\
                               (var_cond1 <= local_config['high_edge']))]

        var_cond2 = df[((local_config['cond2']) &\
                        (local_config['common_cond']))][ivar].values
        var_cond2 = var_cond2[((var_cond2 >= local_config['low_edge']) &\
                               (var_cond2 <= local_config['high_edge']))]

        fig, (ax1, ax2) = plt.subplots(nrows=2, figsize=(8, 6))

        fig.suptitle(r'%s Distribution Comparision - %s | %s' %(local_config['var_name'],
                                                                local_config['title_tag1'],
                                                                local_config['title_tag2']),
                    fontsize=20)

        ns, bins, patches =  ax1.hist([var_cond1, var_cond2],
                                    bins=local_config['nbins'],
                                    color=['blue', 'red'],
                                    histtype='step',
                                    alpha=.5,
                                    label=[local_config['cond1_label'],
                                           local_config['cond2_label']],
                                    lw=2)
        ax1.legend(loc='upper left')

        # ratio plot - vulgo puchadinho
        ax2.plot(bins[:-1],     # this is what makes it comparable
                100*(ns[1] / ns[0]), # maybe check for div-by-zero!
                '^',
                color='black')


        ax1.set_ylabel('Counts', fontsize=15)
        ax2.set_ylabel('Ratio (%)', fontsize=15)
        ax1.set_yscale('log')
        ax2.set_yscale('log')
        ax2.set_xlabel(r'%s' %(local_config['var_name']), fontsize=15)
        # at this time we don't need this
        # ax1.set_xticks(np.arange(start=-0.05, stop=0.06, step=1e-2))
        # ax2.set_xticks(np.arange(start=-0.05, stop=0.06, step=1e-2))
        ax1.set_xlim([local_config['low_edge'], local_config['high_edge']])
        ax2.set_xlim([local_config['low_edge'], local_config['high_edge']])
        ax2.grid()
        ax1.grid()
        plt.savefig(os.path.join(output_path,
                                '%s_mini_quad.png' %(ivar)),
                    dpi=300)





#
# Plot 2D histogram function based on ROOT used by fit_table class
#
def plot_correction( h2, slope, offset, x_points, y_points, error_points, outname, 
                     xlabel='',
                     etBinIdx=None, 
                     etaBinIdx=None, 
                     etBins=None,
                     etaBins=None, 
                     label='Internal',
                     ref_value=None, 
                     pd_value=None,
                     palette=kBlackBody):

    def toStrBin(etlist = None, etalist = None, etidx = None, etaidx = None):
        if etlist and etidx is not None:
            etlist=copy(etlist)
            if etlist[-1]>9999:  etlist[-1]='#infty'
            binEt = (str(etlist[etidx]) + ' < E_{T} [GeV] < ' + str(etlist[etidx+1]) if etidx+1 < len(etlist) else
                                     'E_{T} > ' + str(etlist[etidx]) + ' GeV')
            return binEt
        if etalist and etaidx is not None:
            binEta = (str(etalist[etaidx]) + ' < |#eta| < ' + str(etalist[etaidx+1]) if etaidx+1 < len(etalist) else
                                        str(etalist[etaidx]) + ' <|#eta| < 2.47')
            return binEta

    canvas = TCanvas("canvas","canvas",500,500)
    rpl.set_figure(canvas)
    gStyle.SetPalette(palette)
    canvas.SetRightMargin(0.15)
    canvas.SetTopMargin(0.15)
    canvas.SetLogz()
    h2.GetXaxis().SetTitle('Neural Network output (Discriminant)')
    h2.GetYaxis().SetTitle(xlabel)
    h2.GetZaxis().SetTitle('Count')
    h2.Draw('colz')
    g = TGraphErrors(len(x_points), array.array('d',x_points), array.array('d',y_points), array.array('d',error_points), array.array('d',[0]*len(x_points)))
    g.SetMarkerColor(kBlue)
    g.SetMarkerStyle(8)
    g.SetMarkerSize(1)
    g.Draw("P same")
    line = TLine(slope*h2.GetYaxis().GetXmin()+offset,h2.GetYaxis().GetXmin(), slope*h2.GetYaxis().GetXmax()+offset, h2.GetYaxis().GetXmax())
    line.SetLineColor(kBlack)
    line.SetLineWidth(2)
    line.Draw()
    # Add text labels into the canvas
    text = toStrBin(etlist=etBins, etidx=etBinIdx)
    text+= ', '+toStrBin(etalist=etaBins, etaidx=etaBinIdx)
    if ref_value and pd_value:
        text+=', P_{D} = %1.2f (%1.2f) [%%]'%(pd_value*100, ref_value*100)
    rpl.add_text(0.15, 0.885, text, textsize=.03)
    rpl.set_atlas_label(0.15, 0.94, label)
    rpl.format_canvas_axes(XLabelSize=16, YLabelSize=16, XTitleOffset=0.87, ZLabelSize=16,ZTitleSize=16, YTitleOffset=0.87, ZTitleOffset=1.1)
    canvas.SaveAs(outname)




#
# DISCLAIMER: You should have root installed at your system to use it.
#

__all__ = ['crossval_fit_table']

from Gaugi import Logger, LoggingLevel, StoreGate, restoreStoreGate
from Gaugi.macros import *
from Gaugi import load as gload

from tqdm import tqdm
#from tqdm import tqdm_notebook as tqdm


from copy import copy
from itertools import product
from pprint import pprint

import collections
import numpy as np
import pandas as pd
import time
import array
import json
import collections
import os
import rootplotlib as rpl

from kolmov.utils.plot_functions import plot_correction


import ROOT
from ROOT import kBird,kBlackBody,kTRUE
from ROOT import TCanvas, gStyle, TLegend, gPad, TLatex, TEnv, gROOT, TLine, TColor
from ROOT import kAzure, kRed, kBlue, kBlack,kBird, kOrange,kGray
from ROOT import TGraphErrors,TF1,TH1F,TH2F


from pybeamer import *
from tensorflow.keras.models import Model, model_from_json


rpl.set_atlas_style()
rpl.suppress_root_warnings()
gROOT.SetBatch(kTRUE)

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf


def my_model_generator( row, remove_last=True):
    file_name = row.file_name.values[0]
    tuned_idx = row.tuned_idx.values[0]
    model_idx = row.model_idx.values[0]
    tuned = gload(file_name)['tunedData'][tuned_idx]
    model = model_from_json( json.dumps(tuned['sequence'], separators=(',', ':')) ) #custom_objects={'RpLayer':RpLayer} )
    model.set_weights( tuned['weights'] )
    if remove_last:
        model = Model(model.inputs, model.layers[-2].output)
    return model, file_name, tuned_idx, model_idx

   


#
# Linear correction class table
#
class crossval_fit_table(Logger):

    #
    # Constructor
    #
    def __init__(self, etbins, etabins, kf, data_generator, input_generator,
                 model_generator=my_model_generator):

        # init base class
        Logger.__init__(self)
        self.etbins = etbins
        self.etabins = etabins
        self.__data_generator = data_generator
        self.__input_generator = input_generator
        self.__kf = kf
        self.__model_generator = model_generator




    def fill( self, best_inits, paths, ref_values, xbin_size=0.05, ybin_size=0.5, 
              ymin=16, ymax=60, xmin_percentage = 1, xmax_percentage=99, output_path='', 
              min_avgmu=16, max_avgmu=100, false_alarm_limit=0.5):
        
        tf.config.run_functions_eagerly(False)

        output_path += '/figures'
        try:
            if not os.path.exists(output_path): os.makedirs(output_path)
        except:
            MSG_WARNING( self,'The director %s exist.', output_path)

        columns = best_inits.columns.values.tolist()
        columns.extend([ 
                          'slope',
                          'offset', 
                          'fig_signal', 
                          'fig_background',
                        ])

        dataframe = collections.OrderedDict({col:[] for col in columns})


        for et_bin in best_inits.et_bin.unique():

            for eta_bin in best_inits.eta_bin.unique():

                best_inits_bin = best_inits.loc[(best_inits.et_bin==et_bin)&(best_inits.eta_bin==eta_bin)]

                # check if the current table has this phase space in
                if ( best_inits_bin.shape[0] == 0) :
                    continue

                ref = ref_values[et_bin][eta_bin]

                # read the phase space data to the memory
                data_df = self.__data_generator( paths[et_bin][eta_bin] )
                splits = [(train_idx, val_idx) for train_idx,val_idx in self.__kf.split(data_df, data_df.target)]

                # Loop over all sorts
                for sort in tqdm(best_inits_bin.sort.unique(), 
                                 desc='Running (et=%d, eta%d)'%(et_bin,eta_bin) ):

                    train_idx = splits[sort][0]
                    val_idx   = splits[sort][1]

                    # get model
                    model, file_name, tuned_idx, model_idx = self.__model_generator(best_inits_bin.loc[best_inits_bin.sort==sort])
                    output = model.predict( self.__input_generator(data_df), batch_size=4096).flatten()
                    avgmu  = data_df.avgmu.values.flatten()
                    target = data_df.target.values.flatten()

                    # prepare all histograms
                    xmin = int(np.percentile(output , xmin_percentage))
                    xmax = int(np.percentile(output , xmax_percentage))

                    xbins = int((xmax-xmin)/xbin_size)
                    ybins = int((ymax-ymin)/ybin_size)

                    df_train = pd.DataFrame({ 'output' : output[train_idx], 
                                              'avgmu'  : avgmu[train_idx],
                                              'target' : target[train_idx]})     

                    # Fill train set
                    hist_signal = rpl.hist2d.make_hist('signal', 
                                                        df_train.loc[df_train.target==1].output.values, 
                                                        df_train.loc[df_train.target==1].avgmu.values, 
                                                        xbins, xmin, xmax, ybins, ymin, ymax)

                    hist_background = rpl.hist2d.make_hist('background',
                                                           df_train.loc[df_train.target!=1].output.values, 
                                                           df_train.loc[df_train.target!=1].avgmu.values, 
                                                           xbins, xmin, xmax, ybins, ymin, ymax)

                    if min_avgmu:
                        avgmu[avgmu < min_avgmu] = min_avgmu
                    if max_avgmu:
                        avgmu[avgmu > max_avgmu] = max_avgmu


                    df_train['avgmu'] = avgmu[train_idx]

                    df_val = pd.DataFrame({ 'output':output[val_idx], 
                                            'avgmu' :avgmu[val_idx],
                                            'target':target[val_idx]})

                    df_op  = pd.DataFrame({ 'output':output, 
                                            'avgmu' :avgmu,
                                            'target':target})

                    # Loop over all operation points
                    for op_name in best_inits_bin.op_name.unique():
                    
                        row = best_inits_bin.loc[ (best_inits_bin.sort==sort) &
                                                  (best_inits_bin.op_name==op_name) ]
                        init = row.init.values[0]
                        file_name = row.file_name.values[0] 

                        pd_ref = ref[op_name]['pd']; fa_ref = ref[op_name]['fa']
                        sp_ref  = np.sqrt(  np.sqrt(pd_ref*(1-fa_ref)) * (0.5*(pd_ref+(1-fa_ref)))  )

                        train_tag      = row.train_tag.values[0]
                        max_sp_val     = row.max_sp_val.values[0]
                        max_sp_pd_val  = row.max_sp_pd_val.values[0]
                        max_sp_fa_val  = row.max_sp_fa_val.values[0]
                        max_sp_op      = row.max_sp_op.values[0]
                        max_sp_pd_op   = row.max_sp_pd_op.values[0]
                        max_sp_fa_op   = row.max_sp_fa_op.values[0]
                        pd_ref_passed  = row.pd_ref_passed.values[0]
                        fa_ref_passed  = row.fa_ref_passed.values[0]
                        pd_ref_total   = row.pd_ref_total.values[0]
                        fa_ref_total   = row.fa_ref_total.values[0]

                        #
                        # Train events (calculate slope and offset)
                        #

                        # calculate slope and offset from the train set
                        slope, offset, info = self.calculate( hist_signal, hist_background, pd_ref, false_alarm_limit=false_alarm_limit )
                        df_train['dec'] = np.greater(df_train.output, slope*df_train.avgmu + offset)
                        pd_train_passed = df_train.loc[(df_train.target==True) & (df_train.dec==True)].shape[0]
                        pd_train_total  = df_train.loc[(df_train.target==True)].shape[0]
                        pd_train        = pd_train_passed/pd_train_total

                        #
                        # Validation
                        #

                        df_val['dec'] = np.greater(df_val.output.values, slope*df_val.avgmu.values + offset)
                        # signal (val)
                        pd_val_passed = df_val.loc[(df_val.target==True) & (df_val.dec==True)].shape[0]
                        pd_val_total  = df_val.loc[(df_val.target==True)].shape[0]
                        pd_val = pd_val_passed/pd_val_total

                        # background (val)
                        fa_val_passed = df_val.loc[(df_val.target==False) & (df_val.dec==True)].shape[0]
                        fa_val_total  = df_val.loc[(df_val.target==False)].shape[0]
                        fa_val = fa_val_passed/fa_val_total
                        sp_val = np.sqrt(  np.sqrt(pd_val*(1-fa_val)) * (0.5*(pd_val+(1-fa_val)))  )

                        #
                        # Operation (Train + val)
                        #

                        df_op['dec'] = np.greater(df_op.output, slope*df_op.avgmu + offset)
                        # signal (train)
                        pd_op_passed = df_op.loc[(df_op.target==True) & (df_op.dec==True)].shape[0]
                        pd_op_total  = df_op.loc[(df_op.target==True)].shape[0]
                        pd_op  = pd_op_passed/pd_op_total
                        # background (train)
                        fa_op_passed = df_op.loc[(df_op.target==False) & (df_op.dec==True)].shape[0]
                        fa_op_total  = df_op.loc[(df_op.target==False)].shape[0]
                        fa_op  = fa_op_passed/fa_op_total
                        sp_op  = np.sqrt(  np.sqrt(pd_op*(1-fa_op)) * (0.5*(pd_op+(1-fa_op)))  )

                        #
                        # Plot histograms
                        #

                        label='Internal, #it{%s}'%op_name

                        fig_signal = output_path+'/correction_signal_%s_et%d_eta%d_sort%d.pdf'%(op_name,et_bin,eta_bin,sort)
                        plot_correction( hist_signal, slope, offset, info['x_dots'], info['y_dots'], info['errors'], 
                                         fig_signal, xlabel='<#mu>',
                                         etBinIdx=et_bin, etaBinIdx=eta_bin, 
                                         etBins=self.etbins,etaBins=self.etabins,
                                         label=label, ref_value=pd_ref, 
                                         pd_value=pd_train)

                        fig_background = output_path+'/correction_background_%s_et%d_eta%d_sort%d.pdf'%(op_name,et_bin,eta_bin,sort)
                        plot_correction( hist_background, slope, offset, info['x_dots'], info['y_dots'], info['errors'], 
                                         fig_background, xlabel='<#mu>',
                                         etBinIdx=et_bin, etaBinIdx=eta_bin, 
                                         etBins=self.etbins,etaBins=self.etabins,
                                         label=label, ref_value=pd_ref, 
                                         pd_value=pd_train)


                        # hold all values into the table
                        for col in columns:
                            dataframe[col].append(eval(col))


        self.__table =  pd.DataFrame(dataframe)
        return self.__table

    #
    # Get table
    #
    def table(self):
        return self.__table

    #
    # Fill correction table
    #
    def calculate( self, hist_signal, hist_background, pd_ref, false_alarm_limit = 0.5):

        d = {'x_dots' : [], 'y_dots':[], 'errors':[]}

        false_alarm = 1.0
        while false_alarm > false_alarm_limit:

            # Get the threshold when we not apply any linear correction
            threshold, _ = self.find_threshold( hist_signal.ProjectionX(), pd_ref )
            # Get the efficiency without linear adjustment
            signal_eff, signal_num, signal_den = self.calculate_num_and_den_from_hist(hist_signal, 0.0, threshold)
            
            # Apply the linear adjustment and fix it in case of positive slope
            slope, offset, x_dots, y_dots, errors = self.fit( hist_signal, pd_ref )

            if slope > 0:
                slope = 0; offset = threshold

            # Get the efficiency with linear adjustment
            signal_corrected_eff, signal_corrected_num, signal_corrected_den = self.calculate_num_and_den_from_hist(hist_signal, slope, offset)
            false_alarm, background_corrected_num, background_corrected_den = self.calculate_num_and_den_from_hist(hist_background, slope, offset)

            if false_alarm > false_alarm_limit:
                # Reduce the reference value by hand
                pd_ref-=0.0025


            d['x_dots'] = x_dots
            d['y_dots'] = y_dots
            d['errors'] = errors
        
        return slope, offset, d


    #
    # open the tuned model from npz file
    #

    #
    # Find the threshold given a reference value
    #
    def find_threshold(self, th1,effref):
        nbins = th1.GetNbinsX()
        fullArea = th1.Integral(0,nbins+1)
        if fullArea == 0:
            return 0,1
        notDetected = 0.0; i = 0
        while (1. - notDetected > effref):
            cutArea = th1.Integral(0,i)
            i+=1
            prevNotDetected = notDetected
            notDetected = cutArea/fullArea
        eff = 1. - notDetected
        prevEff = 1. -prevNotDetected
        deltaEff = (eff - prevEff)
        threshold = th1.GetBinCenter(i-1)+(effref-prevEff)/deltaEff*(th1.GetBinCenter(i)-th1.GetBinCenter(i-1))
        error = 1./np.sqrt(fullArea)
        return threshold, error

    #
    # Get all points in the 2D histogram given a reference value
    #
    def get_points( self, th2 , effref):
        nbinsy = th2.GetNbinsY()
        x = list(); y = list(); errors = list()
        for by in range(nbinsy):
            xproj = th2.ProjectionX('xproj'+str(time.time()),by+1,by+1)
            discr, error = self.find_threshold(xproj,effref)
            dbin = xproj.FindBin(discr)
            x.append(discr); y.append(th2.GetYaxis().GetBinCenter(by+1))
            errors.append( error )
        return x,y,errors



    #
    # Calculate the linear fit given a 2D histogram and reference value and return the slope and offset
    #
    def fit(self, th2,effref):
        x_points, y_points, error_points = self.get_points(th2, effref )
        import array
        from ROOT import TGraphErrors, TF1
        g = TGraphErrors( len(x_points)
                             , array.array('d',y_points,)
                             , array.array('d',x_points)
                             , array.array('d',[0.]*len(x_points))
                             , array.array('d',error_points) )
        firstBinVal = th2.GetYaxis().GetBinLowEdge(th2.GetYaxis().GetFirst())
        lastBinVal = th2.GetYaxis().GetBinLowEdge(th2.GetYaxis().GetLast()+1)
        f1 = TF1('f1','pol1',firstBinVal, lastBinVal)
        g.Fit(f1,"FRq")
        slope = f1.GetParameter(1)
        offset = f1.GetParameter(0)
        return slope, offset, x_points, y_points, error_points



    #
    # Calculate the numerator and denomitator given the 2D histogram and slope/offset parameters
    #
    def calculate_num_and_den_from_hist(self, th2, slope, offset) :

      nbinsy = th2.GetNbinsY()
      th1_num = th2.ProjectionY(th2.GetName()+'_proj'+str(time.time()),1,1)
      th1_num.Reset("ICESM")
      numerator=0; denominator=0
      # Calculate how many events passed by the threshold
      for by in range(nbinsy) :
          xproj = th2.ProjectionX('xproj'+str(time.time()),by+1,by+1)
          # Apply the correction using ax+b formula
          threshold = slope*th2.GetYaxis().GetBinCenter(by+1)+ offset
          dbin = xproj.FindBin(threshold)
          num = xproj.Integral(dbin+1,xproj.GetNbinsX()+1)
          th1_num.SetBinContent(by+1,num)
          numerator+=num
          denominator+=xproj.Integral(-1, xproj.GetNbinsX()+1)

      return numerator/denominator, numerator, denominator


    #
    # Convert the table to csv
    #
    def to_csv( self, output ):
        '''
        This function will save the pandas Dataframe into a csv file.

        Arguments.:

        - output: the path and the name to be use for save the table.

        Ex.:
        m_path = './my_awsome_path
        m_name = 'my_awsome_name.csv'

        output = os.path.join(m_path, m_name)
        '''
        self.__table.to_csv(output, index=False)


    #
    # Read the table from csv
    #
    def from_csv( self, input ):
        '''
        This method is used to read a csv file insted to fill the Dataframe from tuned file.

        Arguments:

        - input: the csv file to be opened;
        '''
        self.__table = pd.read_csv(input)
        return self.__table


    #
    # Export all models to keras/onnx using the prometheus format
    #
    def export( self, best_sorts, model_output_format , conf_output, reference_name, to_onnx=False,
                max_avgmu = 100, min_avgmu = 0, remove_last=True):


        try:
            os.makedirs('models')
        except:
            pass

        model_etmin_vec = []
        model_etmax_vec = []
        model_etamin_vec = []
        model_etamax_vec = []
        model_paths = []
        slopes = []
        offsets = []


        for et_bin in best_sorts.et_bin.unique():

            for eta_bin in best_sorts.eta_bin.unique():

                row = best_sorts.loc[(best_sorts.et_bin==et_bin) & (best_sorts.eta_bin==eta_bin)]
                if row.shape[0] == 0:
                    continue

                model, _, _, _ = self.__model_generator( row, remove_last=remove_last)


                model_etmin_vec.append( float(self.etbins[et_bin]) )
                model_etmax_vec.append( float(self.etbins[et_bin+1]) )
                model_etamin_vec.append( float(self.etabins[eta_bin]) )
                model_etamax_vec.append( float(self.etabins[eta_bin+1]) )

                model_name = 'models/'+model_output_format%( et_bin, eta_bin )
                model_paths.append( model_name+'.onnx' ) #default is onnx since this will be used into the athena base
                model_json = model.to_json()

                with open(model_name+".json", "w") as json_file:
                    json_file.write(model_json)
                    # saving the model weight separately
                    model.save_weights(model_name+".h5")

                if to_onnx:
                    # NOTE: This is a hack since I am not be able to convert to onnx inside this function. I need to
                    # open a new python instance (call by system) to convert my models.
                    command = 'convert2onnx.py -j {FILE}.json -w {FILE}.h5 -o {FILE}.onnx'.format(FILE=model_name)
                    os.system(command)


                slopes.append( float(row.slope.values[0]) )
                offsets.append( float(row.offset.values[0]) )


        def list_to_str( l ):
            s = str()
            for ll in l:
              s+=str(ll)+'; '
            return s[:-2]

        # Write the config file
        file = TEnv( 'ringer' )
        file.SetValue( "__name__", 'should_be_filled' )
        file.SetValue( "__version__", 'should_be_filled' )
        file.SetValue( "__operation__", reference_name )
        file.SetValue( "__signature__", 'should_be_filled' )
        file.SetValue( "Model__size"  , str(len(model_paths)) )
        file.SetValue( "Model__etmin" , list_to_str(model_etmin_vec) )
        file.SetValue( "Model__etmax" , list_to_str(model_etmax_vec) )
        file.SetValue( "Model__etamin", list_to_str(model_etamin_vec) )
        file.SetValue( "Model__etamax", list_to_str(model_etamax_vec) )
        file.SetValue( "Model__path"  , list_to_str( model_paths ) )
        file.SetValue( "Threshold__size"  , str(len(model_paths)) )
        file.SetValue( "Threshold__etmin" , list_to_str(model_etmin_vec) )
        file.SetValue( "Threshold__etmax" , list_to_str(model_etmax_vec) )
        file.SetValue( "Threshold__etamin", list_to_str(model_etamin_vec) )
        file.SetValue( "Threshold__etamax", list_to_str(model_etamax_vec) )
        file.SetValue( "Threshold__slope" , list_to_str(slopes) )
        file.SetValue( "Threshold__offset", list_to_str(offsets) )
        file.SetValue( "Threshold__MaxAverageMu", list_to_str([max_avgmu]*len(model_paths)))
        file.SetValue( "Threshold__MinAverageMu", list_to_str([min_avgmu]*len(model_paths)))

        MSG_INFO( self, "Export all tuning configuration to %s.", conf_output)
        file.WriteFile(conf_output)




__all__ = [
    'numpy_to_df'
]

import os
import json

import numpy as np
import pandas as pd

from itertools import product
from tensorflow.keras.models import model_from_json
from Gaugi import Logger
from Gaugi.macros import *
from Gaugi import load as gload



def weights_list_to_array(w_list):
    '''
    A function to transform the weigths from list to array.
    '''
    n_list = []
    for iweigths in w_list:
        n_list.append(np.array(iweigths))
    return n_list



#
# Class numpy_to_df
#
class numpy_to_df(Logger):

    def __init__(self, raw_data):
        '''
        This class has the objective of transform the ringer data into pandas Dataframe
        in order to make easer to manipulate and extract information.

        Arguments:
        - raw_data: the file .npz created in prometheus framework
        Ex.:
        my_data_path =data18_path  some_path_to_data/meaninful_data_name_et%i_eta%i.npz')
        et, eta      = 2, 0
        my_raw_data = dict(np.load(my_data_path %(et, eta)))
        '''
        Logger.__init__(self)
        self.raw_data      = raw_data
        self.et_bin        = raw_data['etBinIdx'].tolist()
        self.eta_bin       = raw_data['etaBinIdx'].tolist()
        # create a pandas DataFrame
        MSG_INFO(self, 'Creating a pandas Dataframe... ')
        self.df_           = pd.DataFrame(data=self.raw_data['data'],
                                          columns=self.raw_data['features'])
        self.df_['target'] = self.raw_data['target']

    def get_df(self):
        '''
        This method return the pandas Dataframe.
        '''
        return self.df_

    def get_rings(self, noHAD=False):
        '''
        This method will return a pandas Dataframe with the rings.
        The flag noHAD is used to remove the hadronic rings.

        Arguments:
        - noHAD: a flag to remove or not the hadronic rings.
        '''
        # decide if want hadronic rings or not
        if noHAD:
            return self.df_[['L2Calo_ring_%i' %(i) for i in range(88)]].copy().values
        else:
            return self.df_[['L2Calo_ring_%i' %(i) for i in range(100)]].copy().values

    def add_tuning_decision(self, tuned_models_dict, jpsiee=True):
        '''
        This method will add into the Dataframe the decision from a given tuning.

        The tuned_models_dict can be more the one tuning. The key in the dictionary
        will be the key of the decision in the Dataframe.

        Arguments:
        - tuned_models_dict: the dictionary which contains the tunings that you want add
        the decision.
        Ex.:
        my_tuned_modes = {
            'model_number1' : (path_to_tuning, path_to_threshold)
        }
        - jpsiee: use the jpsiee et and eta bins
        '''
        if jpsiee:
            aux_index =  list(product(range(3), range(5))).index((self.et_bin, self.eta_bin))
        else:
            aux_index =  list(product(range(5), range(5))).index((self.et_bin, self.eta_bin))
        # FIXME: need to adapt the code to load the new exported tunings
        # loop over the items to build a model dict
        for ikey, (ituning_file, ithr_file) in tuned_models_dict.items():
            MSG_INFO(self, 'Adding %s model to Dataframe' %(ikey))
            with open(ituning_file) as f1, open(ithr_file) as f2:
                tuning_, thr_ = json.load(f1), json.load(f2)
            # get the exact model
            sequential, weights = (tuning_['models'][aux_index]['sequential'],
                                   weights_list_to_array(tuning_['models'][aux_index]['weights']))

            m_thr    = thr_['thresholds'][aux_index]['threshold']
            if m_thr[0] != 0.0:
                print('Not implemented yet!')
                break
            else:
                thr = m_thr[-1]
                print('For %s using threshold: %1.4f \n' %(ikey,thr))
            # build the model
            local_model    = model_from_json(json.dumps(sequential))
            local_model.set_weights(weights)
            print(local_model.summary())
            print('\n')
            # get rings
            rings          = self.get_rings()
            rings         /= np.abs(rings.sum(axis=1))[:, None] # normalization step
            local_output   = local_model.predict(rings)
            # add output
            self.df_[ikey+'_output'] = local_output
            # create the descision
            local_output[ local_output >= thr ] = 1.
            local_output[ local_output < thr ] = 0.
            # add the decision as a new column into the Dataframe
            self.df_[ikey+'_decision'] = local_output





__all__ = ['crossval_table']

from Gaugi.macros import *
from Gaugi import Logger, expand_folders, load, progressbar
from pybeamer import *
from pprint import pprint
from functools import reduce
from itertools import product
from tqdm import tqdm

import matplotlib as mpl
import matplotlib.pyplot as plt
import collections, os, glob, json, copy, re
import numpy as np
import pandas as pd
import os
import json
import joblib
import tensorflow as tf
model_from_json = tf.keras.models.model_from_json



class crossval_table( Logger ):
    #
    # Constructor
    #
    def __init__(self, config_dict, etbins=None, etabins=None ):
        '''
        The objective of this class is extract the tuning information from saphyra's output and
        create a pandas DataFrame using then.
        The informations used in this DataFrame are listed in info_dict, but the user can add more
        information from saphyra summary for example.


        Arguments:

        - config_dict: a dictionary contains in keys the measures that user want to check and
        the values need to be a empty list.

        Ex.: info = collections.OrderedDict( {

              "max_sp_val"      : 'summary/max_sp_val',
              "max_sp_pd_val"   : 'summary/max_sp_pd_val#0',
              "max_sp_fa_val"   : 'summary/max_sp_fa_val#0',
              "max_sp_op"       : 'summary/max_sp_op',
              "max_sp_pd_op"    : 'summary/max_sp_pd_op#0',
              "max_sp_fa_op"    : 'summary/max_sp_fa_op#0',
              'tight_pd_ref'    : "reference/tight_cutbased/pd_ref#0",
              'tight_fa_ref'    : "reference/tight_cutbased/fa_ref#0",
              'tight_pd_ref_passed'     : "reference/tight_cutbased/pd_ref#1",
              'tight_fa_ref_passed'     : "reference/tight_cutbased/fa_ref#1",
              'tight_pd_ref_total'      : "reference/tight_cutbased/pd_ref#2",
              'tight_fa_ref_total'      : "reference/tight_cutbased/fa_ref#2",
              'tight_pd_val_passed'     : "reference/tight_cutbased/pd_val#1",
              'tight_fa_val_passed'     : "reference/tight_cutbased/fa_val#1",
              'tight_pd_val_total'      : "reference/tight_cutbased/pd_val#2",
              'tight_fa_val_total'      : "reference/tight_cutbased/fa_val#2",
              'tight_pd_op_passed'      : "reference/tight_cutbased/pd_op#1",
              'tight_fa_op_passed'      : "reference/tight_cutbased/fa_op#1",
              'tight_pd_op_total'       : "reference/tight_cutbased/pd_op#2",
              'tight_fa_op_total'       : "reference/tight_cutbased/fa_op#2",

              } )

        - etbins: a list of et bins edges used in training;
        - etabins: a list of eta bins edges used in training;
        '''
        Logger.__init__(self)
        self.__table = None
        # Check wanted key type
        self.__config_dict = collections.OrderedDict(config_dict) if type(config_dict) is dict else config_dict
        self.__etbins = etbins
        self.__etabins = etabins


    #
    # Fill the main dataframe with values from the tuning files and convert to pandas dataframe
    #
    def fill(self, path, tag):
        '''
        This method will fill the information dictionary and convert then into a pandas DataFrame.

        Arguments.:

        - path: the path to the tuned files;
        - tag: the training tag used;
        '''
        paths = expand_folders( path )
        MSG_INFO(self, "Reading file for %s tag from %s", tag , path)

        # Creating the dataframe
        dataframe = collections.OrderedDict({
                              'train_tag'      : [],
                              'et_bin'         : [],
                              'eta_bin'        : [],
                              'model_idx'      : [],
                              'sort'           : [],
                              'init'           : [],
                              'file_name'      : [],
                              'tuned_idx'      : [],
                              'op_name'        : [],
                          })

        MSG_INFO(self, 'There are %i files for this task...' %(len(paths)))
        MSG_INFO(self, 'Filling the table... ')

        for ituned_file_name in progressbar( paths , 'Reading %s...'%tag):
            #for ituned_file_name in paths:

            try:
                gfile = load(ituned_file_name)
            except:
                #MSG_WARNING(self, "File %s not open. skip.", ituned_file_name)
                continue
            tuned_file = gfile['tunedData']

            for idx, ituned in enumerate(tuned_file):

                history = ituned['history']
                
                for op, config_dict in self.__config_dict.items():

                    # get the basic from model
                    dataframe['train_tag'].append(tag)
                    dataframe['model_idx'].append(ituned['imodel'])
                    dataframe['sort'].append(ituned['sort'])
                    dataframe['init'].append(ituned['init'])
                    dataframe['et_bin'].append(self.get_etbin(ituned_file_name))
                    dataframe['eta_bin'].append(self.get_etabin(ituned_file_name))
                    dataframe['file_name'].append(ituned_file_name)
                    dataframe['tuned_idx'].append( idx )
                    dataframe['op_name'].append(op)

                    # Get the value for each wanted key passed by the user in the contructor args.
                    for key, local  in config_dict.items():
                        if not key in dataframe.keys():
                            dataframe[key] = [self.__get_value( history, local )]
                        else:
                            dataframe[key].append( self.__get_value( history, local ) )

        # append tables if is need
        # ignoring index to avoid duplicated entries in dataframe
        self.__table = self.__table.append( pd.DataFrame(dataframe), ignore_index=True ) if not self.__table is None else pd.DataFrame(dataframe)
        MSG_INFO(self, 'End of fill step, a pandas DataFrame was created...')


    #
    # Convert the table to csv
    #
    def to_csv( self, output ):
        '''
        This function will save the pandas Dataframe into a csv file.

        Arguments.:

        - output: the path and the name to be use for save the table.

        Ex.:
        m_path = './my_awsome_path
        m_name = 'my_awsome_name.csv'

        output = os.path.join(m_path, m_name)
        '''
        self.__table.to_csv(output, index=False)


    #
    # Read the table from csv
    #
    def from_csv( self, input ):
        '''
        This method is used to read a csv file insted to fill the Dataframe from tuned file.

        Arguments:

        - input: the csv file to be opened;
        '''
        self.__table = pd.read_csv(input)


    #
    # Get the main table
    #
    def table(self):
        '''
        This method will return the pandas Dataframe.
        '''
        return self.__table


    #
    # Set the main table
    #
    def set_table(self, table):
        '''
        This method will set a table indo the main class table.

        Arguments:

        - table: a pandas Dataframe;
        '''
        self.__table=table


    #
    # Get the value using recursive dictionary navigation
    #
    def __get_value(self, history, local):
        '''
        This method will return a value given a history and dictionary with keys.

        Arguments:

        - history: the tuned information file;
        - local: the path caming from config_dict;
        '''
        # Protection to not override the history since this is a 'mutable' object
        var = copy.copy(history)
        for key in local.split('/'):
            var = var[key.split('#')[0]][int(key.split('#')[1])] if '#' in key else var[key]
        return var


    def get_etbin(self, job):
        return int(  re.findall(r'et[a]?[0-9]', job)[0][-1] )


    def get_etabin(self, job):
        return int( re.findall(r'et[a]?[0-9]',job)[1] [-1] )


    def get_etbin_edges(self, et_bin):
        return (self.__etbins[et_bin], self.__etbins[et_bin+1])


    def get_etabin_edges(self, eta_bin):
        return (self.__etabins[eta_bin], self.__etabins[eta_bin+1])


    #
    # Get the pandas dataframe
    #
    def table(self):
        '''
        This method will return the pandas Dataframe.
        '''
        return self.__table


    #
    # Return only best inits
    #
    def filter_inits(self, key, idxmin=False):
        '''
        This method will filter the Dataframe based on given key in order to get the best inits for every sort.

        Arguments:

        - key: the column to be used for filter.
        '''
        if idxmin:
            idxmask = self.table().groupby(['et_bin', 'eta_bin', 'train_tag', 'model_idx', 'sort', 'op_name'])[key].idxmin().values
            return self.table().loc[idxmask]
        else:
            idxmask = self.table().groupby(['et_bin', 'eta_bin', 'train_tag', 'model_idx', 'sort', 'op_name'])[key].idxmax().values
            return self.table().loc[idxmask]


    #
    # Get the best sorts from best inits table
    #
    def filter_sorts(self, best_inits, key, idxmin=False):
        '''
        This method will filter the Dataframe based on given key in order to get the best model for every configuration.

        Arguments:

        - key: the column to be used for filter.
        '''
        if idxmin:
            idxmask = best_inits.groupby(['et_bin', 'eta_bin', 'train_tag', 'model_idx', 'op_name'])[key].idxmin().values
            return best_inits.loc[idxmask]
        else:
            idxmask = best_inits.groupby(['et_bin', 'eta_bin', 'train_tag', 'model_idx', 'op_name'])[key].idxmax().values
            return best_inits.loc[idxmask]


    #
    # Calculate the mean/std table from best inits table
    #
    def describe(cls, best_inits ):
        '''
        This method will give the mean and std for construct the beamer presentation for each train tag.

        Arguments:

        - best_inits:
        '''
        # Create a new dataframe to hold this table
        dataframe = { 'train_tag' : [], 'et_bin' : [], 'eta_bin' : [], 'op_name':[],
                      'pd_ref':[], 'fa_ref':[], 'sp_ref':[]}
        # Include all wanted keys into the dataframe
        for key in best_inits.columns.values:
            if key in ['train_tag', 'et_bin', 'eta_bin', 'op_name']:
                continue
            elif 'passed' in key or 'total' in key:
                continue
            elif ('op' in key) or ('val' in key):
                dataframe[key+'_mean'] = []; dataframe[key+'_std'] = []
            else:
                continue

        # Loop over all tuning tags and et/eta bins
        for tag in best_inits.train_tag.unique():
            for et_bin in best_inits.et_bin.unique():
                for eta_bin in best_inits.eta_bin.unique():
                    for op in best_inits.op_name.unique():

                        cv_bin = best_inits.loc[ (best_inits.train_tag == tag) & (best_inits.et_bin == et_bin) & (best_inits.eta_bin == eta_bin) & (best_inits.op_name == op)]
                        dataframe['train_tag'].append( tag )
                        dataframe['et_bin'].append( et_bin )
                        dataframe['eta_bin'].append( eta_bin )
                        dataframe['op_name'].append(op)
                        dataframe['pd_ref'].append(cv_bin.pd_ref.values[0])
                        dataframe['fa_ref'].append(cv_bin.fa_ref.values[0])
                        dataframe['sp_ref'].append(cv_bin.sp_ref.values[0])

                        for key in best_inits.columns.values:
                            if key in ['train_tag', 'et_bin', 'eta_bin', 'op_name']:
                                continue # skip common
                            elif 'passed' in key or 'total' in key:
                                continue # skip counts
                            elif ('op' in key) or ('val' in key):
                                dataframe[key+'_mean'].append( cv_bin[key].mean() ); dataframe[key+'_std'].append( cv_bin[key].std() )
                            else: # skip others
                                continue


        # Return the pandas dataframe
        return pd.DataFrame(dataframe)



    #
    # Get tge cross val integrated table from best inits
    #
    def integrate( self, best_inits, tag ):
        '''
        This method is used to get the integrate information of a given tag.

        Arguments:

        - best_inits: a pandas Dataframe which contains all information for the best inits.
        - tag: the training tag that will be integrate.
        '''
        keys = [ key for key in best_inits.columns.values if 'passed' in key or 'total' in key]
        table = best_inits.loc[best_inits.train_tag==tag].groupby(['sort']).agg(dict(zip( keys, ['sum']*len(keys))))

        for key in keys:
            if 'passed' in key:
                orig_key = key.replace('_passed','')
                values = table[key].div( table[orig_key+'_total'] )
                table[orig_key] = values
                table=table.drop([key],axis=1)
                table=table.drop([orig_key+'_total'],axis=1)
        return table.agg(['mean','std'])


    def evaluate( self, best_sorts, paths , data_generator, dec_generator ):

        tf.config.run_functions_eagerly(False)

        columns = best_sorts.columns.values.tolist()
        extra = ['pd_test', 'fa_test', 'sp_test', 'pd_test_passed', 'pd_test_total', 'fa_test_passed', 'fa_test_total']
        columns.extend(extra)
        table = collections.OrderedDict({ key:[] for key in columns} )

        bins = list(product(range(len(self.__etbins)-1),range(len(self.__etabins)-1)))

        # Loop over all et/eta bins
        for et_bin, eta_bin in tqdm( bins , desc= 'Fitting... ', ncols=70):

                data = data_generator(paths[et_bin][eta_bin])

                for train_tag in best_sorts.train_tag.unique():


                    rows = best_sorts.loc[(best_sorts.et_bin==et_bin) & (best_sorts.eta_bin==eta_bin) & (best_sorts.train_tag==train_tag) ] 
                    
                    for op_name in rows.op_name.unique():

                        row = rows.loc[rows.op_name==op_name]
                        data['dec'] = dec_generator( row, data )

                        pd_test_passed = data.loc[(data.target==1) & (data.dec==True)].shape[0]
                        pd_test_total = data.loc[(data.target==1)].shape[0]
                        pd_test = pd_test_passed/pd_test_total

                        fa_test_passed = data.loc[(data.target!=1) & (data.dec==True)].shape[0]
                        fa_test_total = data.loc[(data.target!=1)].shape[0]
                        fa_test = fa_test_passed/fa_test_total

                        sp_test = np.sqrt(  np.sqrt(pd_test*(1-fa_test)) * (0.5*(pd_test+(1-fa_test)))  )

                        for col in columns:
                            if col in extra:
                                continue
                            table[col].append( getattr(row, col).values[0] )
                        
                        for col in extra:
                            table[col].append(eval(col))

        return pd.DataFrame(table)



    #
    # Dump all history for each line in the table
    #
    def dump_all_history( self, table, output_path , tag):
        '''
        This method will dump the train history. This is a way to get more easy this information when plotting the train evolution.

        Arguments:

        - table: a table with the path information.
        - output_path: the path to sabe the hitories.
        - tag: the train tag.
        '''
        if not os.path.exists( output_path ):
          os.mkdir( output_path )
        for _ , row in table.iterrows():
            if row.train_tag != tag:
              continue
            # Load history
            history = self.get_history(row.file_name, row.tuned_idx)
            history['loc'] = {'et_bin' : row.et_bin, 'eta_bin' : row.eta_bin, 'sort' : row.sort, 'model_idx' : row.model_idx}
            name = 'history_et_%i_eta_%i_model_%i_sort_%i.json' % (row.et_bin,row.eta_bin,row.model_idx,row.sort)
            jbl_name = 'history_et_%i_eta_%i_model_%i_sort_%i.joblib' % (row.et_bin,row.eta_bin,row.model_idx,row.sort)
            joblib.dump(history['summary'], os.path.join(output_path, jbl_name))
            history.pop('summary')
            with open(os.path.join(output_path, '%s' %name), 'w', encoding='utf-8') as fp:
                #json.dump(transform_serialize(history), fp)
                json.dump(str(history), fp)



    def get_history( self, path, index ):
      tuned_list = load(path)['tunedData']
      for tuned in tuned_list:
        if tuned['imodel'] == index:
          return tuned['history']
      MSG_FATAL( self, "It's not possible to find the history for model index %d", index )



	#
	# Plot the training curves for all sorts.
	#
    def plot_training_curves( self, best_inits, best_sorts, dirname, display=False, start_epoch=3 ):
        '''
        This method is a shortcut to plot the monitoring traning curves.

        Arguments:

        - best_inits: a pandas Dataframe which contains all information for the best inits;
        - best_sorts: a pandas Dataframe which contains all information for the best sorts;
        - dirname: a folder to save the figures, if not exist we'll create and attached in $PWD folder;
        - display: a boolean to decide if show or not show the plot;
        - start_epoch: the epoch to start draw the plot.
        '''
        basepath = os.getcwd()
        if not os.path.exists(basepath+'/'+dirname):
          os.mkdir(basepath+'/'+dirname)

        def plot_training_curves_for_each_sort(table, et_bin, eta_bin, best_sort , output, display=False, start_epoch=0):
            table = table.loc[(table.et_bin==et_bin) & (table.eta_bin==eta_bin)]
            nsorts = len(table.sort.unique())
            fig, ax = plt.subplots(nsorts,2, figsize=(15,20))
            fig.suptitle(r'Monitoring Train Plot - Et = %d, Eta = %d'%(et_bin,eta_bin), fontsize=15)
            for idx, sort in enumerate(table.sort.unique()):
                current_table = table.loc[table.sort==sort]
                path=current_table.file_name.values[0]
                history = self.get_history( path, current_table.model_idx.values[0])
                
                best_epoch = history['max_sp_best_epoch_val'][-1] - start_epoch
                # Make the plot here
                ax[idx, 0].set_xlabel('Epochs')
                ax[idx, 0].set_ylabel('Loss (sort = %d)'%sort, color = 'r' if best_sort==sort else 'k')
                ax[idx, 0].plot(history['loss'][start_epoch::], c='b', label='Train Step')
                ax[idx, 0].plot(history['val_loss'][start_epoch::], c='r', label='Validation Step')
                ax[idx, 0].axvline(x=best_epoch, c='k', label='Best epoch')
                ax[idx, 0].legend()
                ax[idx, 0].grid()
                ax[idx, 1].set_xlabel('Epochs')
                ax[idx, 1].set_ylabel('SP (sort = %d)'%sort, color = 'r' if best_sort==sort else 'k')
                ax[idx, 1].plot(history['max_sp_val'][start_epoch::], c='r', label='Validation Step')
                ax[idx, 1].axvline(x=best_epoch, c='k', label='Best epoch')
                ax[idx, 1].legend()
                ax[idx, 1].grid()

            plt.savefig(output)
            if display:
                plt.show()
            else:
                plt.close(fig)

        for tag in best_inits.train_tag.unique():
            for et_bin in best_inits.et_bin.unique():
                for eta_bin in best_inits.eta_bin.unique():
                    best_sort = best_sorts.loc[(best_sorts.et_bin==et_bin) & (best_sorts.eta_bin==eta_bin) & (best_sorts.train_tag==tag)].sort 
                    plot_training_curves_for_each_sort(best_inits.loc[best_inits.train_tag==tag], et_bin, eta_bin, best_sort.values[0],
                        basepath+'/'+dirname+'/train_evolution_et%d_eta%d_%s.pdf'%(et_bin,eta_bin,tag), display, start_epoch)



    #
    # Plot the training curves for all sorts.
    #
    def plot_roc_curves( self, best_sorts, tags, legends, output, display=False, colors=None, points=None, et_bin=None, eta_bin=None,
                         xmin=-0.02, xmax=0.3, ymin=0.8, ymax=1.02, fontsize=18, figsize=(15,15)):
        '''
        This method will plot the ROC curves.

        Arguments:

        - best_sorts: a pandas Dataframe which contains all information for the best sorts;
        - tag: the tuning tags to be plotted
        - legends: the legends to used (unused)
        - output: the path with the name to save the plot;
        - display: a boolean to decide if show or not show the plot;
        - colors: the color list to use in plot;
        - points: the points list to diferenciate the curves (unused);
        - et_bin: the et bin index
        - eta_bin: the eta bin index
        - xmin: x plot lower limit;
        - xmax: x plot upper limit;
        - ymin: y plot lower limit;
        - ymax: y plot upper limit;
        - fontsize: font size just like in matplotlib
        - figsize: figure size just like in matplotlib
        '''

        def plot_roc_curves_for_each_bin(ax, table, colors, xmin=-0.02, xmax=0.3, ymin=0.8, ymax=1.02, fontsize=18):

          ax.set_xlabel('Fake Probability [%]',fontsize=fontsize)
          ax.set_ylabel('Detection Probability [%]',fontsize=fontsize)
          ax.set_title(r'Roc curve (et = %d, eta = %d)'%(table.et_bin.values[0], table.eta_bin.values[0]),fontsize=fontsize)

          for idx, tag in enumerate(tags):
              current_table = table.loc[(table.train_tag==tag)]

              path=current_table.file_name.values[0]
              history = self.get_history( path, current_table.model_idx.values[0])
              pd, fa = history['summary']['rocs']['roc_op']
              ax.plot( fa, pd, color=colors[idx], linewidth=2, label=tag)
              ax.set_ylim(ymin,ymax)
              ax.set_xlim(xmin,xmax)

          ax.legend(fontsize=fontsize)
          ax.grid()


        if et_bin!=None and eta_bin!=None:
            fig, ax = plt.subplots(1,1, figsize=figsize)
            fig.suptitle(r'Operation ROCs', fontsize=15)
            table_for_this_bin = best_sorts.loc[(best_sorts.et_bin==et_bin) & (best_sorts.eta_bin==eta_bin)]
            plot_roc_curves_for_each_bin( ax, table_for_this_bin, colors, xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, fontsize=fontsize)
            plt.savefig(output)
            if display:
                plt.show()
            else:
                plt.close(fig)

        else:

            n_et_bins = len(best_sorts.et_bin.unique())
            n_eta_bins = len(best_sorts.eta_bin.unique())
            fig, ax = plt.subplots(n_et_bins,n_eta_bins, figsize=figsize)
            fig.suptitle(r'Operation ROCs', fontsize=15)

            for et_bin in best_sorts.et_bin.unique():
                for eta_bin in best_sorts.eta_bin.unique():
                    table_for_this_bin = best_sorts.loc[(best_sorts.et_bin==et_bin) & (best_sorts.eta_bin==eta_bin)]
                    plot_roc_curves_for_each_bin( ax[et_bin][eta_bin], table_for_this_bin, colors, xmin=xmin, xmax=xmax,
                                                  ymin=ymin, ymax=ymax, fontsize=fontsize)

            plt.savefig(output)
            if display:
                plt.show()
            else:
                plt.close(fig)




    #
    # Create the beamer table file
    #
    def dump_beamer_table( self, best_inits, output, tags=None, title='' , test_table=None):
        '''
        This method will use a pandas Dataframe in order to create a beamer presentation which summary the tuning cross-validation.

        Arguments:
        - best_inits: a pandas Dataframe which contains all information for the best inits.
        - operation_poins: the operation point that will be used.
        - output: a name for the pdf
        - tags: the training tag that will be used. If None then the tags will be get from the Dataframe.
        - title: the pdf title
        '''

        cv_table = self.describe( best_inits )
        # Create Latex Et bins
        etbins_str = []; etabins_str = []
        for etBinIdx in range(len(self.__etbins)-1):
            etbin = (self.__etbins[etBinIdx], self.__etbins[etBinIdx+1])
            if etbin[1] > 100 :
                etbins_str.append( r'$E_{T}\text{[GeV]} > %d$' % etbin[0])
            else:
                etbins_str.append(  r'$%d < E_{T} \text{[Gev]}<%d$'%etbin )

        # Create Latex eta bins
        for etaBinIdx in range( len(self.__etabins)-1 ):
            etabin = (self.__etabins[etaBinIdx], self.__etabins[etaBinIdx+1])
            etabins_str.append( r'$%.2f<\eta<%.2f$'%etabin )

        # Default colors
        colorPD = '\\cellcolor[HTML]{9AFF99}'; colorPF = ''; colorSP = ''

        train_tags = cv_table.train_tag.unique() if not tags else tags

        # fix tags to list
        if type(tags) is str: tags=[tags]

        # Apply beamer
        with BeamerTexReportTemplate1( theme = 'Berlin'
                                 , _toPDF = True
                                 , title = title
                                 , outputFile = output
                                 , font = 'structurebold' ):

                ### Prepare tables
                tuning_names = ['']; tuning_names.extend( train_tags )
                lines1 = []
                lines1 += [ HLine(_contextManaged = False) ]
                lines1 += [ HLine(_contextManaged = False) ]
                lines1 += [ TableLine( columns = ['','','kinematic region'] + reduce(lambda x,y: x+y,[['',s,''] for s in etbins_str]), _contextManaged = False ) ]
                lines1 += [ HLine(_contextManaged = False) ]
                lines1 += [ TableLine( columns = ['Det. Region','Method','Type'] + reduce(lambda x,y: x+y,[[colorPD+r'$P_{D}[\%]$',colorSP+r'$SP[\%]$',colorPF+r'$P_{F}[\%]$'] \
                                      for _ in etbins_str]), _contextManaged = False ) ]
                lines1 += [ HLine(_contextManaged = False) ]

                for etaBinIdx in cv_table.eta_bin.unique():
                    for idx, tag in enumerate( train_tags ):
                        cv_values=[]; ref_values=[]; test_values=[]
                        for etBinIdx in cv_table.et_bin.unique():
                            current_table = cv_table.loc[ (cv_table.train_tag==tag) & (cv_table.et_bin==etBinIdx) & (cv_table.eta_bin==etaBinIdx) ]
                            sp = current_table['sp_val_mean'].values[0]*100
                            pd = current_table['pd_val_mean'].values[0]*100
                            fa = current_table['fa_val_mean'].values[0]*100
                            sp_std = current_table['sp_val_std'].values[0]*100
                            pd_std = current_table['pd_val_std'].values[0]*100
                            fa_std = current_table['fa_val_std'].values[0]*100
                            sp_ref = current_table['sp_ref'].values[0]*100
                            pd_ref = current_table['pd_ref'].values[0]*100
                            fa_ref = current_table['fa_ref'].values[0]*100

                            cv_values   += [ colorPD+('%1.2f$\pm$%1.2f')%(pd,pd_std),colorSP+('%1.2f$\pm$%1.2f')%(sp,sp_std),colorPF+('%1.2f$\pm$%1.2f')%(fa,fa_std),    ]
                            ref_values  += [ colorPD+('%1.2f')%(pd_ref), colorSP+('%1.2f')%(sp_ref), colorPF+('%1.2f')%(fa_ref)]
                            
                            if test_table is not None:
                                current_test_table = test_table.loc[ (test_table.train_tag==tag) & (test_table.et_bin==etBinIdx) & (test_table.eta_bin==etaBinIdx) ]
                                pd = current_test_table['pd_test'].values[0]*100
                                sp = current_test_table['sp_test'].values[0]*100
                                fa = current_test_table['fa_test'].values[0]*100
                                test_values   += [ colorPD+('%1.2f')%(pd),colorSP+('%1.2f')%(sp),colorPF+('%1.2f')%(fa)    ]

                        if idx > 0:
                            lines1 += [ TableLine( columns = ['', tuning_names[idx+1], 'Cross Val.'] + cv_values   , _contextManaged = False ) ]
                        
                        
                            if test_table is not None:
                                lines1 += [ TableLine( columns = ['', tuning_names[idx+1], 'Test'] + test_values   , _contextManaged = False ) ]
                    
                        
                        else:
                            lines1 += [ TableLine( columns = ['\multirow{%d}{*}{'%(len(tuning_names))+etabins_str[etaBinIdx]+'}',tuning_names[idx], 'Reference'] + ref_values   ,
                                                _contextManaged = False ) ]
                            lines1 += [ TableLine( columns = ['', tuning_names[idx+1], 'Cross Val.'] + cv_values    , _contextManaged = False ) ]
                            if test_table is not None:
                                lines1 += [ TableLine( columns = ['', tuning_names[idx+1], 'Test'] + test_values   , _contextManaged = False ) ]
                    
                    lines1 += [ HLine(_contextManaged = False) ]
                lines1 += [ HLine(_contextManaged = False) ]


                ### Calculate the final efficiencies
                lines2 = []
                lines2 += [ HLine(_contextManaged = False) ]
                lines2 += [ HLine(_contextManaged = False) ]
                lines2 += [ TableLine( columns = ['',colorPD+r'$P_{D}[\%]$',colorPF+r'$F_{a}[\%]$'], _contextManaged = False ) ]
                lines2 += [ HLine(_contextManaged = False) ]
                for idx, tag in enumerate( train_tags ):
                    current_table = self.integrate( best_inits, tag )
                    pd = current_table['pd_val'].values[0]*100
                    pd_std = current_table['pd_val'].values[1]*100
                    fa = current_table['fa_val'].values[0]*100
                    fa_std = current_table['fa_val'].values[1]*100

                    pd_ref = current_table['pd_ref'].values[0]*100
                    fa_ref = current_table['fa_ref'].values[0]*100

                    if test_table is not None:
                        keys = [ key for key in test_table.columns.values if 'passed' in key or 'total' in key]
                        current_test_table = test_table.loc[test_table.train_tag=='v8'].agg(dict(zip( keys, ['sum']*len(keys))))
                        test_pd = (current_test_table['pd_test_passed']/current_test_table['pd_test_total'])*100
                        test_fa = (current_test_table['fa_test_passed']/current_test_table['fa_test_total'])*100

                    if idx > 0:
                        lines2 += [ TableLine( columns = [tag.replace('_','\_') + ' (Cross Val.)' ,
                          colorPD+('%1.2f$\pm$%1.2f')%(pd,pd_std),colorPF+('%1.2f$\pm$%1.2f')%(fa,fa_std) ], _contextManaged = False ) ]
                    
                        if test_table is not None:
                            lines2 += [ TableLine( columns = [tag.replace('_','\_') + ' (Test)',
                            colorPD+('%1.2f')%(test_pd),colorPF+('%1.2f')%(test_fa) ], _contextManaged = False ) ]
                    
                    else:
                        lines2 += [ TableLine( columns = ['Ref.' ,colorPD+('%1.2f')%(pd_ref),colorPF+('%1.2f')%(fa_ref) ], _contextManaged = False ) ]
                        lines2 += [ TableLine( columns = [tag.replace('_','\_') + ' (Cross Val.)',
                                    colorPD+('%1.2f$\pm$%1.2f')%(pd,pd_std),colorPF+('%1.2f$\pm$%1.2f')%(fa,fa_std) ], _contextManaged = False ) ]
                        if test_table is not None:
                            lines2 += [ TableLine( columns = [tag.replace('_','\_')+ ' (Test)' ,
                            colorPD+('%1.2f')%(test_pd),colorPF+('%1.2f')%(test_fa) ], _contextManaged = False ) ]
                    

                # Create all tables into the PDF Latex
                with BeamerSlide( title = "The Cross Validation Efficiency Values For All Tunings"  ):
                    with Table( caption = 'The $P_{d}$, $F_{a}$ and $SP $ values for each phase space for each method.') as table:
                        with ResizeBox( size = 1. ) as rb:
                            with Tabular( columns = '|lcc|' + 'ccc|' * len(etbins_str) ) as tabular:
                                tabular = tabular
                                for line in lines1:
                                    if isinstance(line, TableLine):
                                        tabular += line
                                    else:
                                        TableLine(line, rounding = None)

                with BeamerSlide( title = "The General Efficiency"  ):
                    with Table( caption = 'The general efficiency for the cross validation method for each method.') as table:
                        with ResizeBox( size = 0.7 ) as rb:
                            with Tabular( columns = 'lc' + 'c' * 2 ) as tabular:
                                tabular = tabular
                                for line in lines2:
                                    if isinstance(line, TableLine):
                                        tabular += line
                                    else:
                                        TableLine(line, rounding = None)



    #
    # Load all keras models given the best sort table
    #
    def get_best_models( self, best_sorts , remove_last=True, with_history=True):
        '''
        This method will load the best models.

        Arguments:

        - best_sorts: the table that contains the best_sorts;
        - remove_last: a bolean variable to remove or not the tanh in tha output layer;
        - with_history: unused variable.
        '''
        from tensorflow.keras.models import Model, model_from_json
        import json

        models = [[ None for _ in range(len(self.__etabins)-1)] for __ in range(len(self.__etbins)-1)]
        for et_bin in range(len(self.__etbins)-1):
            for eta_bin in range(len(self.__etabins)-1):
                d_tuned = {}
                best = best_sorts.loc[(best_sorts.et_bin==et_bin) & (best_sorts.eta_bin==eta_bin)]
                tuned = load(best.file_name.values[0])['tunedData'][best.tuned_idx.values[0]]
                model = model_from_json( json.dumps(tuned['sequence'], separators=(',', ':')) ) #custom_objects={'RpLayer':RpLayer} )
                model.set_weights( tuned['weights'] )
                new_model = Model(model.inputs, model.layers[-2].output) if remove_last else model
                #new_model.summary()
                d_tuned['model']    = new_model
                d_tuned['etBin']    = [self.__etbins[et_bin], self.__etbins[et_bin+1]]
                d_tuned['etaBin']   = [self.__etabins[eta_bin], self.__etabins[eta_bin+1]]
                d_tuned['etBinIdx'] = et_bin
                d_tuned['etaBinIdx']= eta_bin
                d_tuned['sort']     = best.sort.values[0]
                d_tuned['init']     = best.init.values[0]
                d_tuned['model_idx']= best.model_idx.values[0]
                d_tuned['file_name']= best.file_name.values[0]
                if with_history:
                    d_tuned['history']  = tuned['history']
                models[et_bin][eta_bin] = d_tuned
        return models





    #
    # Load all keras models given the best sort table
    #
    def get_best_init_models( self, best_inits , remove_last=True, with_history=True):
        '''
        This method will load the best init models.

        Arguments:

        - best_inits: the table that contains the best_inits;
        - remove_last: a bolean variable to remove or not the tanh in tha output layer;
        - with_history: unused variable.
        '''
        from tensorflow.keras.models import Model, model_from_json
        import json

        models = [[ [] for _ in range(len(self.__etabins)-1)] for __ in range(len(self.__etbins)-1)]

        for et_bin in range(len(self.__etabins)-1):
            for eta_bin in range(len(self.__etabins)-1):
                for sort in best_inits.sort.unique():
                    d_tuned = {}
                    best = best_inits.loc[(best_inits.et_bin==et_bin) & (best_inits.eta_bin==eta_bin) & (best_inits.sort==sort)]
                    tuned = load(best.file_name.values[0])['tunedData'][best.tuned_idx.values[0]]
                    model = model_from_json( json.dumps(tuned['sequence'], separators=(',', ':')) ) #custom_objects={'RpLayer':RpLayer} )
                    model.set_weights( tuned['weights'] )
                    new_model = Model(model.inputs, model.layers[-2].output) if remove_last else model
                    #new_model.summary()
                    d_tuned['model']    = new_model
                    d_tuned['etBin']    = [self.__etbins[et_bin], self.__etbins[et_bin+1]]
                    d_tuned['etaBin']   = [self.__etabins[eta_bin], self.__etabins[eta_bin+1]]
                    d_tuned['etBinIdx'] = et_bin
                    d_tuned['etaBinIdx']= eta_bin
                    d_tuned['sort']     = best.sort.values[0]
                    d_tuned['init']     = best.init.values[0]
                    d_tuned['model_idx']= best.model_idx.values[0]
                    d_tuned['file_name']= best.file_name.values[0]
                    if with_history:
                        d_tuned['history']  = tuned['history']
                    models[et_bin][eta_bin].append(d_tuned)
        return models



__all__ = [
    'first_export_tool',
    'export_tool',
    'export_onnx_tool',
    'export_fastnet_to_onnx'
]

import os
import json

import numpy as np
import pandas as pd

from itertools import product
from Gaugi import load as gload
from Gaugi import save as gsave
from sklearn.metrics import roc_curve
from tensorflow.keras.models import model_from_json

from kolmov.utils.constants import etbins_zee, etbins_jpsiee, etabins

import onnx
import keras2onnx
from ROOT import TEnv, TFile

# to get the legacy correctly
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Activation


# SP index definition
def calc_sp(pd, fa):
    g_mean = np.sqrt((1-fa) * pd)
    a_mean = ((1-fa) + pd)/2.
    return np.sqrt(g_mean*a_mean)

aux_threshold_dict_keys = {
    'Tight'     : 'tight_op_threshold',
    'Medium'    : 'medium_op_threshold',
    'Loose'     : 'loose_op_threshold',
    'VeryLoose' : 'vloose_op_threshold',
}



class export_tool(object):
    '''
    This class is the default kolmov export tool to get and prepare the tunings
    to move them for prometheus framework.
    '''

    def __init__(self, operation_dataframe):
        '''
        Arguments:
        - operation_dataframe: a .csv file with all models selected to operation.
        '''
        self.op_df            = pd.read_csv(operation_dataframe)


    def get_models_dict(self):
        return self.model_dict

    def get_threshold_dict(self):
        return self.threshold_dict

    def model_finder(self, tunedfile, model_idx, sort, init):
        '''
        This function will search in the tunedfile for wanted
        model given the model idx, sort and init returning the sequential and the weights.

        Arguments:

        - tunedfile: the file with tuned models
        Ex.: 'tunedDiscr.jobID_0004.pic.gz'

        - model_idx: the model index, this is more convinient when you have many MLP's models.
        - sort: the sort that you want.
        - init: the initialization that you want.
        '''
        for itunedData in tunedfile['tunedData']:
            if ((itunedData['imodel'] == model_idx)\
                and (itunedData['sort'] == sort)\
                and (itunedData['init'] == init)):
                # return the keras sequential and weights
                return itunedData['sequence'], itunedData['weights']

    def save_dicts(self, operation_point):
        '''
        This function will save a dictionary into a json file.

        Arguments:
        -dictionary: is the python dict that you want to save;
        -filename: the name of dictionary to be save.
        '''
        thr_filename = 'TrigL2CaloRingerElectron%sThresholds.json' %operation_point
        m_filename   = 'TrigL2CaloRingerElectron%sConstants.json' %operation_point
        with open(thr_filename, 'w') as fp:
            json.dump(self.threshold_dict, fp)
        with open(m_filename, 'w') as fp:
            json.dump(self.model_dict, fp)


    def fill_models_thr_dict(self, operation_point, tuning_tag,
                             tuning_name, isJpsiee=True, save_json=True):
        '''
        This function will fill the dictionary of operation models using the information
        from the operation dataframe and using the ringer data to calculate the threshold.
        '''

        # models dictionary
        self.model_dict = {}
        self.model_dict['models']          = [] # this must be a list
        self.model_dict['__version__']     = tuning_tag
        self.model_dict['__type__']        = 'Model'
        self.model_dict['__name__']        = tuning_name
        self.model_dict['__description__'] = ''
        # threshold dictionary
        self.threshold_dict = {}
        self.threshold_dict['thresholds']      = [] # same as model
        self.threshold_dict['__version__']     = tuning_tag
        self.threshold_dict['__type__']        = 'Threshold'
        self.threshold_dict['__name__']        = tuning_name
        self.threshold_dict['__description__'] = ''

        etabin_list = etabins
        if isJpsiee:
            etbin_list = etbins_jpsiee
        else:
            etbin_list = etbins_zee

        # set the operation label
        self.model_dict['__operation__'] = operation_point
        self.threshold_dict['__operation__'] = operation_point
        thr_dataframe_key = aux_threshold_dict_keys[operation_point]

        for iet, ieta in product(range(self.op_df.et_bin.nunique()),
                                 range(self.op_df.eta_bin.nunique())):
            print('Processing et bin: %i | eta bin: %i' %(iet, ieta))

            # we need a local dictionary
            m_local_dict = {}
            thr_local_dict = {}
            # 1 step: get the right file, model_id, sort and init to open.
            aux_df = self.op_df.loc[((self.op_df['et_bin'] == iet) &\
                                     (self.op_df['eta_bin'] == ieta)), :].copy()
            f_name, id_model, sort, init, thr = aux_df.iloc[0][['file_name', \
                                                                'model_idx', \
                                                                'sort', \
                                                                'init', \
                                                                thr_dataframe_key]].values

            # model info: sequential, weights and binning information
            # add the threshold to thr_local dictionary
            m, w = self.model_finder(tunedfile=gload(f_name),
                                     model_idx=id_model,
                                     sort=sort,
                                     init=init)
            m_local_dict['sequential'] = m
            m_local_dict['weights']    = [wi.tolist() for wi in w] #np.arrays are not serialized
            m_local_dict['etBin']  = etbin_list[iet]
            m_local_dict['etaBin'] = etabin_list[ieta]

            # thr info: threshold and binning information
            # this information must to be a list with 3 itens [alpha, beta, raw_threshold]
            # since wee do not have pile up correction in saphyra the configuration must to be
            # [0., raw_threshold, raw_threshold]
            thr_local_dict['threshold'] = [0., thr, thr]
            thr_local_dict['etBin']  = etbin_list[iet]
            thr_local_dict['etaBin'] = etabin_list[ieta]

            # append the local dict to dict
            self.model_dict['models'].append(m_local_dict)
            self.threshold_dict['thresholds'].append(thr_local_dict)
        if save_json:
            print('Saving json files...')
            self.save_dicts(operation_point)
        print('Done!')

# to export in onnx format

class export_fastnet_to_onnx(object):
    '''
    This class was created to export tunings that were made in TuningTools (FastNet) in order to move them to prometheus framework.
    '''

    def __init__(self, path_to_tuning):
        '''
        Arguments:
        path_to_tunings: the path that contains the .root files
        '''
        self.thr_filename    = os.path.join(path_to_tuning, 'TrigL2CaloRingerElectron%sThresholds.root')
        self.models_filename = os.path.join(path_to_tuning, 'TrigL2CaloRingerElectron%sConstants.root')

    # auxiliar function to create the export files
    def list_to_str(self, l ):
        s = str()
        for ll in l:
            s+=str(ll)+'; '
        return s[:-2]

    def translate_bins(self, et, eta, isJpsiee=True):
        '''
        This function will translate the bins from root file.
        Arguments:
        et: et value from root file
        eta: eta value from root file
        isJpsiee: boolean to specific the et bins.
        '''
        if isJpsiee:
            if et[0] == 0:
                etBin = 0
            elif et[0] == 7:
                etBin = 1
            elif et[0] == 10:
                etBin = 2
        else:
            if et[0] == 0 or et[0] == 15:
                etBin = 0
            elif et[0] == 20:
                etBin = 1
            elif et[0] == 30:
                etBin = 2
            elif et[0] == 40:
                etBin = 3
            elif et[0] == 50:
                etBin = 4
        # get eta
        if eta[0] == 0:
            etaBin = 0
        elif eta[0] == 0.8:
            etaBin = 1
        elif eta[0] == 1.37:
            etaBin = 2
        elif eta[0] == 1.54:
            etaBin = 3
        elif eta[0] == 2.37:
            etaBin = 4
        return etBin, etaBin

    def convert_to_keras_model(self, neuron, W, B, input_shape=100 ):
        '''
        This function will create a keras model which will be filled with FastNet model information.
        This function works only and only for MLP models with one hidden layer.
        Arguments:
        neuron: #neurons in hidden layer.
        W: FastNet weights.
        B: FastNet bias.
        input_shape: shape of input layer
        '''
        # Build the standard model
        model  = Sequential()
        model.add( Dense( neuron , input_shape= (input_shape,), activation = 'tanh') )
        model.add( Dense( 1, activation='linear' ) )
        w0 = np.zeros((neuron, input_shape))
        b0 = []
        for i in range(neuron):
            for j in range(input_shape):
                w0[i][j] = W.pop(0)
            b0.append(B.pop(0))
        w1 = W
        b1 = B
        ww = np.array([ np.array(w0).T, np.array(b0), np.array(w1), np.array(b1) ] )
        ww[2]=ww[2].reshape((ww[2].shape[0],1))
        model.set_weights(ww)
        return model

    def create_config_files(self, operation_point, tuning_name, version,
                            model_format_name, output_config_name,
                            signature='electron', isJpsiee=True):
        '''
        This function will fill the dictionary of operation models using the information
        from the operation dataframe and using the ringer data to calculate the threshold.
        '''

        etabin_list = etabins
        if isJpsiee:
            etbin_list = etbins_jpsiee
        else:
            etbin_list = etbins_zee

        # format the file names, open then and get the tree
        thrs_file      = TFile(self.thr_filename %(operation_point))
        thresholdsTree = thrs_file.tuning.Get('thresholds')

        models_file    = TFile(self.models_filename %(operation_point))
        modelsTree     = models_file.tuning.Get('discriminators')

        # tuning information
        model_etmin_vec = []
        model_etmax_vec = []
        model_etamin_vec = []
        model_etamax_vec = []
        model_paths = []

        # slopes and offsets (threshold configuration)
        slopes  = []
        offsets = []
        # loop over models and thresholds in TTree
        for ithr, imodel in zip(thresholdsTree, modelsTree):
            # get the bins
            iet, ieta = self.translate_bins(ithr.etBin, ithr.etaBin)
            print('Processing et bin: %i | eta bin: %i' %(iet, ieta))

            # fill et and eta vec
            model_etmin_vec.append(etbin_list[iet][0])
            model_etmax_vec.append(etbin_list[iet][1])
            model_etamin_vec.append(etabin_list[ieta][0])
            model_etamax_vec.append(etabin_list[ieta][1])

            # easy way to get the thresholds
            l_thr   = list(ithr.thresholds)
            slopes.append(l_thr[0])
            offsets.append(l_thr[1])

            # get the models
            weights = list(imodel.weights)
            bias = list(imodel.bias)
            l_model = self.convert_to_keras_model(len(weights)%100, weights, bias)
            print(l_model.summary())
            # convert keras to Onnx
            onnx_model = keras2onnx.convert_keras(l_model, l_model.name)
            # onnx model name and add to the model paths
            s_l = model_format_name %(operation_point, iet, ieta)
            # now add the model string to model_paths
            onnx_model_name = s_l
            model_paths.append( s_l.split('/')[-1] )
            # save the onnx model
            onnx.save_model(onnx_model, onnx_model_name)
        # done! Fill all list with the models
        # write the config file
        m_file = TEnv('ringer')
        m_file.SetValue("__name__", tuning_name)
        m_file.SetValue("__version__", version)
        m_file.SetValue("__operation__", operation_point)
        m_file.SetValue("__signature__", signature)
        m_file.SetValue("Model__size", str(len(model_paths)))
        m_file.SetValue("Model__etmin", self.list_to_str(model_etmin_vec))
        m_file.SetValue("Model__etmax", self.list_to_str(model_etmax_vec))
        m_file.SetValue("Model__etamin", self.list_to_str(model_etamin_vec))
        m_file.SetValue("Model__etamax", self.list_to_str(model_etamax_vec))
        m_file.SetValue("Model__path", self.list_to_str(model_paths))
        m_file.SetValue( "Threshold__size"  , str(len(model_paths)) )
        m_file.SetValue( "Threshold__etmin" , self.list_to_str(model_etmin_vec) )
        m_file.SetValue( "Threshold__etmax" , self.list_to_str(model_etmax_vec) )
        m_file.SetValue( "Threshold__etamin", self.list_to_str(model_etamin_vec) )
        m_file.SetValue( "Threshold__etamax", self.list_to_str(model_etamax_vec) )
        m_file.SetValue( "Threshold__slope" , self.list_to_str(slopes) )
        m_file.SetValue( "Threshold__offset", self.list_to_str(offsets) )
        m_file.SetValue( "Threshold__MaxAverageMu", 100.)
        m_file.WriteFile(output_config_name%(operation_point))



class export_onnx_tool(object):
    '''
    This class is the default kolmov export tool to get and prepare the tunings
    to move them for prometheus framework in onnx format.
    '''

    def __init__(self, operation_dataframe):
        '''
        Arguments:
        - operation_dataframe: a .csv file with all models selected to operation.
        '''
        self.op_df            = pd.read_csv(operation_dataframe)


    def model_finder(self, tunedfile, model_idx, sort, init):
        '''
        This function will search in the tunedfile for wanted
        model given the model idx, sort and init returning the sequential and the weights.

        Arguments:

        - tunedfile: the file with tuned models
        Ex.: 'tunedDiscr.jobID_0004.pic.gz'

        - model_idx: the model index, this is more convinient when you have many MLP's models.
        - sort: the sort that you want.
        - init: the initialization that you want.
        '''
        for itunedData in tunedfile['tunedData']:
            if ((itunedData['imodel'] == model_idx)\
                and (itunedData['sort'] == sort)\
                and (itunedData['init'] == init)):
                # return the keras sequential and weights
                return itunedData['sequence'], itunedData['weights']

    # auxiliar function to create the export files
    def list_to_str(self, l ):
        s = str()
        for ll in l:
            s+=str(ll)+'; '
        return s[:-2]

    def create_config_files(self, operation_point, tuning_name, version,
                            model_format_name, output_config_name,
                            signature='electron', isJpsiee=True):
        '''
        This function will fill the dictionary of operation models using the information
        from the operation dataframe and using the ringer data to calculate the threshold.
        '''

        etabin_list = etabins
        if isJpsiee:
            etbin_list = etbins_jpsiee
        else:
            etbin_list = etbins_zee

        thr_dataframe_key = aux_threshold_dict_keys[operation_point]
        # loop over et and eta
        # tuning information
        model_etmin_vec = []
        model_etmax_vec = []
        model_etamin_vec = []
        model_etamax_vec = []
        model_paths = []
        # slopes and offsets (threshold configuration)
        slopes  = []
        offsets = []
        # loop over et and eta
        for iet, ieta in product(range(self.op_df.et_bin.nunique()),
                                 range(self.op_df.eta_bin.nunique())):
            print('Processing et bin: %i | eta bin: %i' %(iet, ieta))


            # 1 step: get the right file, model_id, sort and init to open.
            aux_df = self.op_df.loc[((self.op_df['et_bin'] == iet) &\
                                     (self.op_df['eta_bin'] == ieta)), :].copy()
            f_name, id_model, sort, init, thr = aux_df.iloc[0][['file_name', \
                                                                'model_idx', \
                                                                'sort', \
                                                                'init', \
                                                                thr_dataframe_key]].values

            # model info: sequential, weights and binning information
            # add the threshold to thr_local dictionary
            m, w = self.model_finder(tunedfile=gload(f_name),
                                     model_idx=id_model,
                                     sort=sort,
                                     init=init)

            # fill et and eta vec
            model_etmin_vec.append(etbin_list[iet][0])
            model_etmax_vec.append(etbin_list[iet][1])
            model_etamin_vec.append(etabin_list[ieta][0])
            model_etamax_vec.append(etabin_list[ieta][1])

            # set the weights to the model
            l_model = model_from_json(json.dumps(m))
            l_model.set_weights(w)
            # remove the tanh in output
            l_model.pop()
            print(l_model.summary())
            # convert keras to Onnx
            onnx_model = keras2onnx.convert_keras(l_model, l_model.name)
            # onnx model name and add to the model paths
            s_l = model_format_name %(operation_point, iet, ieta)

            onnx_model_name = s_l
            model_paths.append( s_l.split('/')[-1] )
            # save the onnx model
            onnx.save_model(onnx_model, onnx_model_name)

            # add dummy thresholds
            slopes.append(0.)
            offsets.append(0.)
        # done! Fill all list with the models
        # write the config file
        m_file = TEnv('ringer')
        m_file.SetValue("__name__", tuning_name)
        m_file.SetValue("__version__", version)
        m_file.SetValue("__operation__", operation_point)
        m_file.SetValue("__signature__", signature)
        m_file.SetValue("Model__size", str(len(model_paths)))
        m_file.SetValue("Model__etmin", self.list_to_str(model_etmin_vec))
        m_file.SetValue("Model__etmax", self.list_to_str(model_etmax_vec))
        m_file.SetValue("Model__etamin", self.list_to_str(model_etamin_vec))
        m_file.SetValue("Model__etamax", self.list_to_str(model_etamax_vec))
        m_file.SetValue("Model__path", self.list_to_str(model_paths))
        m_file.SetValue( "Threshold__size"  , str(len(model_paths)) )
        m_file.SetValue( "Threshold__etmin" , self.list_to_str(model_etmin_vec) )
        m_file.SetValue( "Threshold__etmax" , self.list_to_str(model_etmax_vec) )
        m_file.SetValue( "Threshold__etamin", self.list_to_str(model_etamin_vec) )
        m_file.SetValue( "Threshold__etamax", self.list_to_str(model_etamax_vec) )
        m_file.SetValue( "Threshold__slope" , self.list_to_str(slopes) )
        m_file.SetValue( "Threshold__offset", self.list_to_str(offsets) )
        m_file.SetValue( "Threshold__MaxAverageMu", 100.)
        m_file.WriteFile(output_config_name%(operation_point))

########################
#TOGUIDE
# def convert_to_onnx_with_dummy_thresholds( models, name, version, signature, model_output_format , operation, output):

#     import onnx
#     import keras2onnx
#     from ROOT import TEnv
#     model_etmin_vec = []
#     model_etmax_vec = []
#     model_etamin_vec = []
#     model_etamax_vec = []
#     model_paths = []

#     slopes = []
#     offsets = []

#     for model in models:

#         model_etmin_vec.append( model['etBin'][0] )
#         model_etmax_vec.append( model['etBin'][1] )
#         model_etamin_vec.append( model['etaBin'][0] )
#         model_etamax_vec.append( model['etaBin'][1] )

#         etBinIdx = model['etBinIdx']
#         etaBinIdx = model['etaBinIdx']

#         # Conver keras to Onnx
#         onnx_model = keras2onnx.convert_keras(model['model'], model['model'].name)

#         onnx_model_name = model_output_format%( etBinIdx, etaBinIdx )
#         model_paths.append( onnx_model_name )

#         # Save onnx mode!
#         onnx.save_model(onnx_model, onnx_model_name)

#         slopes.append( 0.0 )
#         offsets.append( 0.0 )


#     def list_to_str( l ):
#         s = str()
#         for ll in l:
#           s+=str(ll)+'; '
#         return s[:-2]

#     # Write the config file
#     file = TEnv( 'ringer' )
#     file.SetValue( "__name__", name )
#     file.SetValue( "__version__", version )
#     file.SetValue( "__operation__", operation )
#     file.SetValue( "__signature__", signature )
#     file.SetValue( "Model__size"  , str(len(models)) )
#     file.SetValue( "Model__etmin" , list_to_str(model_etmin_vec) )
#     file.SetValue( "Model__etmax" , list_to_str(model_etmax_vec) )
#     file.SetValue( "Model__etamin", list_to_str(model_etamin_vec) )
#     file.SetValue( "Model__etamax", list_to_str(model_etamax_vec) )
#     file.SetValue( "Model__path"  , list_to_str( model_paths ) )
#     file.SetValue( "Threshold__size"  , str(len(models)) )
#     file.SetValue( "Threshold__etmin" , list_to_str(model_etmin_vec) )
#     file.SetValue( "Threshold__etmax" , list_to_str(model_etmax_vec) )
#     file.SetValue( "Threshold__etamin", list_to_str(model_etamin_vec) )
#     file.SetValue( "Threshold__etamax", list_to_str(model_etamax_vec) )
#     file.SetValue( "Threshold__slope" , list_to_str(slopes) )
#     file.SetValue( "Threshold__offset", list_to_str(offsets) )
#     file.SetValue( "Threshold__MaxAverageMu", 100)
#     file.WriteFile(output)

# for op in ['Tight','Medium','Loose','VeryLoose']:
#     format = 'data17_13TeV_EGAM1_probes_lhmedium_EGAM7_vetolhvloose.model_v10.electron'+op+'.et%d_eta%d.onnx'
#     output = "ElectronRinger%sTriggerConfig.conf"%op
#     convert_to_onnx_with_dummy_thresholds( models, 'TrigL2_20200715_v10', 'v10', 'electron', format ,op ,output)

################

# this tool will calculate the threshold in case that this info aren't avaliable
class first_export_tool(object):
    '''
    This class is the very first export tool to get and prepare the tunings
    to move them for prometheus framework.

    As in older version of saphyra tunings the thresholds wasn't saved so this class
    will compute the operation threshold using the data used to train.

    This probably will be changed when saphyra dump the thresholds too.
    '''

    def __init__(self, task_path, operation_dataframe, ringer_data_path, noHAD=False):
        '''
        Arguments:
        - task_path: the path where are all tuned files.
        - operation_dataframe: a .csv file with all models selected to operation.
        - ringer_data: path to ringer_data that can be formated the et and eta.
        - noHAD: noHAD flag to use only EM rings.

        Ex.:
        ringer_data = your_data_path + 'your_file_name...et%i_eta%i.npz'

        '''
        self.task_path         = task_path
        self.op_df            = pd.read_csv(operation_dataframe)
        self.ringer_data_path = ringer_data_path
        self.noHAD            = noHAD

    def model_finder(self, tunedfile, model_idx, sort, init):
        '''
        This function will search in the tunedfile for wanted
        model given the model idx, sort and init returning the sequential and the weights.

        Arguments:

        - tunedfile: the file with tuned models
        Ex.: 'tunedDiscr.jobID_0004.pic.gz'

        - model_idx: the model index, this is more convinient when you have many MLP's models.
        - sort: the sort that you want.
        - init: the initialization that you want.
        '''
        for itunedData in tunedfile['tunedData']:
            if ((itunedData['imodel'] == model_idx)\
                and (itunedData['sort'] == sort)\
                and (itunedData['init'] == init)):
                # return the keras sequential and weights
                return itunedData['sequence'], itunedData['weights']

    def save_models_dict(self, filename):
        '''
        This function will save the models dictionary obtained in fill_models_dict function.

        Arguments:
        -filename: the name of dictionary to be save.
        '''
        gsave(self.models, '%s.pic.gz' %(filename))

    def fill_models_dict(self):
        '''
        This function will fill the dictionary of operation models using the information
        from the operation dataframe and using the ringer data to calculate the threshold.
        '''
        self.models = {}
        # add the noHAD flag
        self.models['noHAD'] = self.noHAD
        for iet, ieta in product(range(self.op_df.et_bin.nunique()),
                                 range(self.op_df.eta_bin.nunique())):
            print('Processing et bin: %i | eta bin: %i' %(iet, ieta))
            # alocate in the dictionary
            self.models['et%i_eta%i' %(iet, ieta)] = {}

            # 1 step: get the right file, model_id, sort and init to open.
            aux_df = self.op_df.loc[((self.op_df['et_bin'] == iet) &\
                                     (self.op_df['eta_bin'] == ieta)), :].copy()
            f_name, id_model, sort, init = aux_df.iloc[0][['file_name', \
                                                        'model_idx', \
                                                        'sort', \
                                                        'init']].values

            # 2 step: get the model and the weights
            tuned_file = os.path.join(self.task_path, f_name) %(iet, ieta)
            m, w = self.model_finder(tunedfile=gload(tuned_file), model_idx=id_model,
                                sort=sort, init=init)
            # save the model and the weights into the dictionary
            self.models['et%i_eta%i' %(iet, ieta)]['sequential'] = m
            self.models['et%i_eta%i' %(iet, ieta)]['weights']    = w
            # load the keras models in order to get the thresholds
            local_model = model_from_json(json.dumps(m))
            local_model.set_weights(w)

            # 3 step: load the data to predict using the associated model
            print('Load rings data... ')
            data             = gload(self.ringer_data_path %(iet, ieta))
            if self.noHAD: # check if we need all rings or no...
                print('This is a noHAD tuning, using only EM rings...')
                rings            = data['data'][:, 1:89]
            else:
                rings            = data['data'][:, 1:101]
            print('Preprocessing rings...')
            norm             = np.abs(rings.sum(axis=1))
            norm[ norm == 0] = 1
            rings            = rings/norm[:, None] # normalized rings
            trgt             = data['target']
            # predict on rings
            print('Predict step...')
            nn_output        = local_model.predict(rings)
            print('Getting the operation threshold... ')
            # calculate the thresholds
            fa, pd, thresholds = roc_curve(trgt, nn_output)
            sp = calc_sp(pd, fa)
            # get the knee
            knee = np.argmax(sp)
            # save the threshold
            self.models['et%i_eta%i' %(iet, ieta)]['threshold'] = thresholds[knee]
            print(10*'-')
        print('Done!')







#
# DISCLAIMER: You should have root installed at your system to use it.
#

__all__ = ['fit_table']

from Gaugi import Logger, LoggingLevel, StoreGate, restoreStoreGate
from Gaugi.macros import *
from Gaugi import load as gload

from tqdm import tqdm
from copy import copy
from itertools import product

import numpy as np
import pandas as pd
import time
import array
import json
import collections
import os
import rootplotlib as rpl


import ROOT
from ROOT import kBird,kBlackBody,kTRUE
from ROOT import TCanvas, gStyle, TLegend, gPad, TLatex, TEnv, gROOT, TLine, TColor
from ROOT import kAzure, kRed, kBlue, kBlack,kBird, kOrange,kGray
from ROOT import TGraphErrors,TF1,TH1F,TH2F


from pybeamer import *
from tensorflow.keras.models import Model, model_from_json


rpl.set_atlas_style()
rpl.suppress_root_warnings()
gROOT.SetBatch(kTRUE)

# allow gpu growth 
#os.environ['TF_FORCE_GPU_ALLOW_GROWTH'] = 'true'
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import tensorflow as tf

#import tensorflow as tf
#tf.config.run_functions_eagerly(True)
#tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)


#
# Linear correction class table
#
class fit_table(Logger):

    #
    # Constructor
    #
    def __init__(self, etbins, etabins):

        # init base class
        Logger.__init__(self)
        self.__etbins = etbins
        self.__etabins = etabins
    


    #
    # Fill correction table
    #
    def fill( self, generator, data_paths,  models, output_file, x_bin_size, y_bin_size, ymin, ymax,
              verbose=False, 
              xmin_percentage=1, 
              xmax_percentage=99, 
              xmin=None, 
              xmax=None):

 
        store = StoreGate(output_file)
        bins = list(product(range(len(self.__etbins)-1),range(len(self.__etabins)-1)))

        # Loop over all et/eta bins
        for et_bin, eta_bin in tqdm( bins , desc= 'Filling... ', ncols=70):

            path = data_paths[et_bin][eta_bin]
            data, target, avgmu = generator(path)

            model = models[et_bin][eta_bin]
            # Get the predictions
            tf.config.run_functions_eagerly(False)
            outputs = model['model'].predict(data, batch_size=1024, verbose=verbose).flatten()

            # Get all limits using the output
            xmin = xmin if xmin else int(np.percentile(outputs , xmin_percentage))
            xmax = xmax if xmax else int(np.percentile(outputs , xmax_percentage))

            MSG_DEBUG(self, 'Setting xmin to %1.2f and xmax to %1.2f', xmin, xmax)
            xbins = int((xmax-xmin)/x_bin_size)


            if type(y_bin_size ) is float:
                ybins = int((ymax-ymin)/y_bin_size)

                th2_signal = TH2F( 'th2_signal_et%d_eta%d'%(et_bin,eta_bin), '', xbins, xmin, xmax, ybins, ymin, ymax )
                th2_background = TH2F( 'th2_background_et%d_eta%d'%(et_bin,eta_bin), '', xbins, xmin, xmax, ybins, ymin, ymax )
            
            else:
                y_bins_edges = y_bin_size[et_bin][eta_bin]
                th2_signal = TH2F( 'th2_signal_et%d_eta%d'%(et_bin,eta_bin), '', xbins, xmin, xmax, 
                                 len(y_bins_edges)-1, array.array('d', y_bins_edges))
                th2_background = TH2F( 'th2_background_et%d_eta%d'%(et_bin,eta_bin), '', xbins,xmin, xmax, 
                                        len(y_bins_edges)-1, array.array('d', y_bins_edges) )

            # fill hists
            w = array.array( 'd', np.ones_like( outputs[target==1] ) )
            th2_signal.FillN( len(outputs[target==1]), array.array('d',  outputs[target==1].tolist()),  array.array('d',avgmu[target==1].tolist()), w)
            
            w = array.array( 'd', np.ones_like( outputs[target==0] ) )
            th2_background.FillN( len(outputs[target==0]), array.array('d',outputs[target!=1].tolist()), array.array('d',avgmu[target!=1].tolist()), w)
            
            store.addHistogram(th2_signal)
            store.addHistogram(th2_background)

        store.write()





    #
    # Fill correction table
    #
    def calculate( self, input_file, models, reference_values, output_dir, 
                   except_these_bins = [],
                   pallete = kBlackBody,
                   plot_label = 'Internal',
                   false_alarm_limit = 0.5 ):

        self.__palette = pallete
        # create directory
        localpath = os.getcwd()+'/'+output_dir
        try:
            if not os.path.exists(localpath): os.makedirs(localpath)
        except:
            MSG_WARNING( self,'The director %s exist.', localpath)


        store = restoreStoreGate(input_file)


        # make template dataframe
        dataframe = collections.OrderedDict({
                      'name':[],
                      'et_bin':[],
                      'eta_bin':[],
                      'reference_signal_passed':[],
                      'reference_signal_total':[],
                      'reference_signal_eff':[],
                      'reference_background_passed':[],
                      'reference_background_total':[],
                      'reference_background_eff':[],
                      'signal_passed':[],
                      'signal_total':[],
                      'signal_eff':[],
                      'background_passed':[],
                      'background_total':[],
                      'background_eff':[],
                      'signal_corrected_passed':[],
                      'signal_corrected_total':[],
                      'signal_corrected_eff':[],
                      'background_corrected_passed':[],
                      'background_corrected_total':[],
                      'background_corrected_eff':[],
                     })

        # reduce verbose
        def add(key,value):
          dataframe[key].append(value)

        bins = list(product(range(len(self.__etbins)-1),range(len(self.__etabins)-1)))

        # Loop over all et/eta bins
        for et_bin, eta_bin in tqdm( bins , desc= 'Fitting... ', ncols=70):

            references = reference_values[et_bin][eta_bin]

            model = models[et_bin][eta_bin]
            model['thresholds'] = {}
            th2_signal = store.histogram('th2_signal_et%d_eta%d'%(et_bin,eta_bin))
            th2_background = store.histogram('th2_background_et%d_eta%d'%(et_bin,eta_bin))
            

            MSG_DEBUG( self, 'Applying linear correction to et%d_eta%d bin.', et_bin, eta_bin)

            for name, ref in references.items():
                
                if ref['pd_epsilon'] == 0.0:
                    ref_value = ref['pd']
                else:
                    add_fac = (1-ref['pd'])*ref['pd_epsilon']
                    ref_value = ref['pd'] + add_fac
                    MSG_INFO(self, 'Add %1.2f %% in reference pd -> new reference pd: %1.2f', ref['pd_epsilon'], add_fac)

                false_alarm = 1.0
                while false_alarm > false_alarm_limit:

                    # Get the threshold when we not apply any linear correction
                    threshold, _ = self.find_threshold( th2_signal.ProjectionX(), ref_value )

                    # Get the efficiency without linear adjustment
                    signal_eff, signal_num, signal_den = self.calculate_num_and_den_from_hist(th2_signal, 0.0, threshold)
                    #signal_eff, signal_num, signal_den = self.calculate_num_and_den_from_output(outputs[target==1], avgmu[target==1], 0.0, threshold)
                    background_eff, background_num, background_den = self.calculate_num_and_den_from_hist(th2_background, 0.0, threshold)
                    #background_eff, background_num, background_den = self.calculate_num_and_den_from_output(outputs[target!=1], avgmu[target!=1], 0.0, threshold)

                    # Apply the linear adjustment and fix it in case of positive slope
                    slope, offset, x_points, y_points, error_points = self.fit( th2_signal, ref_value )
                    
                    # put inside of the ref array
                    apply_fit = True

                    # case 1: The user select each bin will not be corrected
                    for (this_et_bin, this_eta_bin) in except_these_bins:
                        if et_bin == this_et_bin and eta_bin == this_eta_bin:
                            apply_fit = False
                    # case 2: positive slope
                    if slope > 0:
                        MSG_WARNING(self, "Retrieved positive angular factor of the linear correction, setting to 0!")
                        apply_fit = False


                    slope = slope if apply_fit else 0
                    offset = offset if apply_fit else threshold

                    # Get the efficiency with linear adjustment
                    signal_corrected_eff, signal_corrected_num, signal_corrected_den = self.calculate_num_and_den_from_hist(th2_signal, slope, offset)
                    #signal_corrected_eff, signal_corrected_num, signal_corrected_den = self.calculate_num_and_den_from_output(outputs[target==1], \
                    #                                                                                                          avgmu[target==1], slope, offset)
                    background_corrected_eff, background_corrected_num, background_corrected_den = self.calculate_num_and_den_from_hist(th2_background, slope, offset)
                    #background_corrected_eff, background_corrected_num, background_corrected_den = self.calculate_num_and_den_from_output(outputs[target!=1], \
                    #                                                                                                                      avgmu[target!=1], slope, offset)

                    false_alarm = background_corrected_num/background_corrected_den # get the passed/total

                    if false_alarm > false_alarm_limit:
                        # Reduce the reference value by hand
                        ref_value-=0.0025

                MSG_DEBUG( self, 'Reference name: %s, target: %1.2f%%', name, ref['pd']*100 )
                MSG_DEBUG( self, 'Signal with correction is: %1.2f%%', signal_corrected_num/signal_corrected_den * 100 )
                MSG_DEBUG( self, 'Background with correction is: %1.2f%%', background_corrected_num/background_corrected_den * 100 )

                # decore the model array
                model['thresholds'][name] = {'offset':offset,
                                             'slope':slope,
                                             'threshold' : threshold,
                                             'reference_pd': ref['pd'],
                                             'reference_fa':ref['fa'],
                                             }
                paths = []

                #
                # Plot histograms
                #

                label = plot_label + ', #it{%s}'%ref['label']
                info = models[et_bin][eta_bin]['thresholds'][name]

                outname = localpath+'/th2_signal_%s_et%d_eta%d.pdf'%(name,et_bin,eta_bin)
                output = self.plot_2d_hist( th2_signal, slope, offset, x_points, y_points, error_points, outname, xlabel='<#mu>',
                                   etBinIdx=et_bin, etaBinIdx=eta_bin, etBins=self.__etbins,etaBins=self.__etabins,
                                   label=label, ref_value=ref['pd'], pd_value=signal_num/signal_den)
                paths.append(output)

                outname = localpath+'/th2_background_%s_et%d_eta%d.pdf'%(name,et_bin,eta_bin)
                output = self.plot_2d_hist( th2_background, slope, offset, x_points, y_points, error_points, outname, xlabel='<#mu>',
                                   etBinIdx=et_bin, etaBinIdx=eta_bin, etBins=self.__etbins,etaBins=self.__etabins,
                                   label=label, ref_value=ref['pd'], pd_value=signal_num/signal_den)
                paths.append(output)

                model['thresholds'][name]['figures'] = paths

                # et/eta bin information
                add( 'name'                        , name )
                add( 'et_bin'                      , et_bin  )
                add( 'eta_bin'                     , eta_bin )
                # reference values
                add( 'reference_signal_passed'     , int(ref['pd']*signal_den) )
                add( 'reference_signal_total'      , signal_den )
                add( 'reference_signal_eff'        , ref['pd'] )
                add( 'reference_background_passed' , int(ref['fa']*background_den) )
                add( 'reference_background_total'  , background_den )
                add( 'reference_background_eff'    , ref['fa'] )
                # non-corrected values
                add( 'signal_passed'               , signal_num )
                add( 'signal_total'                , signal_den )
                add( 'signal_eff'                  , signal_num/signal_den )
                add( 'background_passed'           , background_num )
                add( 'background_total'            , background_den )
                add( 'background_eff'              , background_num/background_den )
                # corrected values
                add( 'signal_corrected_passed'     , signal_corrected_num )
                add( 'signal_corrected_total'      , signal_corrected_den )
                add( 'signal_corrected_eff'        , signal_corrected_num/signal_corrected_den )
                add( 'background_corrected_passed' , background_corrected_num )
                add( 'background_corrected_total'  , background_corrected_den )
                add( 'background_corrected_eff'    , background_corrected_num/background_corrected_den )

        # convert to pandas dataframe
        self.__table = pd.DataFrame( dataframe )
        return self.__table

    #
    # Get the pandas table
    #
    def table(self):
        return self.__table


    #
    # Dump bearmer report
    #
    def dump_beamer_table( self, table, models, title, op_name, output_pdf ):

        name = op_name

        # Slide maker
        with BeamerTexReportTemplate1( theme = 'Berlin'
                                     , _toPDF = True
                                     , title = title
                                     , outputFile = output_pdf
                                     , font = 'structurebold' ):

            # Create Latex Et bins
            etbins_str = []; etabins_str = []
            for etBinIdx in range(len(self.__etbins)-1):
                etbin = (self.__etbins[etBinIdx], self.__etbins[etBinIdx+1])
                if etbin[1] > 100 :
                    etbins_str.append( r'$E_{T}\text{[GeV]} > %d$' % etbin[0])
                else:
                    etbins_str.append(  r'$%d < E_{T} \text{[Gev]}<%d$'%etbin )

            # Create Latex eta bins
            for etaBinIdx in range( len(self.__etabins)-1 ):
                etabin = (self.__etabins[etaBinIdx], self.__etabins[etaBinIdx+1])
                etabins_str.append( r'$%.2f<\eta<%.2f$'%etabin )


            with BeamerSection( name = name.replace('_','\_') ):
                # prepare figures
                with BeamerSubSection( name = 'Correction plots for each phase space'):
                    for etBinIdx in range( len(self.__etbins)-1 ):
                        for etaBinIdx in range( len(self.__etabins)-1 ):
                            current_table = table.loc[ (table.name==name) & (table.et_bin==etBinIdx) & (table.eta_bin==etaBinIdx)]
                            # prepate 2D histograms
                            paths = models[etBinIdx][etaBinIdx]['thresholds'][name]['figures']
                            title = 'Energy between %s in %s (et%d\_eta%d)'%(etbins_str[etBinIdx],
                                                                             etabins_str[etaBinIdx],
                                                                             etBinIdx,etaBinIdx)
                            BeamerMultiFigureSlide( title = title
                            , paths = paths
                            , nDivWidth = 2 # x
                            , nDivHeight = 1 # y
                            , texts=None
                            , fortran = False
                            , usedHeight = 0.6
                            , usedWidth = 1.
                            )
                # prepate table
                with BeamerSubSection( name = 'Efficiency Values' ):
                    # Prepare phase space table
                    lines1 = []
                    lines1 += [ HLine(_contextManaged = False) ]
                    lines1 += [ HLine(_contextManaged = False) ]
                    lines1 += [ TableLine( columns = [''] + [s for s in etabins_str], _contextManaged = False ) ]
                    lines1 += [ HLine(_contextManaged = False) ]
                    for etBinIdx in range( len(self.__etbins)-1 ):
                        values_det = []; values_fa = []
                        for etaBinIdx in range( len(self.__etabins)-1 ):
                            # Get the current bin table
                            current_table = table.loc[ (table.name==name) & (table.et_bin==etBinIdx) & (table.eta_bin==etaBinIdx)]
                            det = current_table.signal_corrected_eff.values[0] * 100
                            fa = current_table.background_corrected_eff.values[0] * 100
                            # Get reference pd
                            ref = current_table.reference_signal_eff.values[0] * 100
                            if (det-ref) > 0.0:
                                values_det.append( ('\\cellcolor[HTML]{9AFF99}%1.2f ($\\uparrow$%1.2f[$\\Delta_{ref}$])')%(det,det-ref) )
                            elif (det-ref) < 0.0:
                                values_det.append( ('\\cellcolor[HTML]{F28871}%1.2f ($\\downarrow$%1.2f[$\\Delta_{ref}$])')%(det,det-ref) )
                            else:
                                values_det.append( ('\\cellcolor[HTML]{9AFF99}%1.2f')%(det) )
                            ref = current_table.reference_background_eff.values[0] * 100
                            factor = fa/ref if ref else 0.
                            if (fa-ref) > 0.0:
                                values_fa.append( ('\\cellcolor[HTML]{F28871}%1.2f ($\\rightarrow$%1.2f$\\times\\text{FR}_{ref}$)')%(fa,factor) )
                            elif (fa-ref) < 0.0:
                                values_fa.append( ('\\cellcolor[HTML]{9AFF99}%1.2f ($\\rightarrow$%1.2f$\\times\\text{FR}_{ref}$)')%(fa,factor) )
                            else:
                                values_fa.append( ('\\cellcolor[HTML]{9AFF99}%1.2f')%(fa) )
                        lines1 += [ TableLine( columns = [etbins_str[etBinIdx]] + values_det   , _contextManaged = False ) ]
                        lines1 += [ TableLine( columns = [''] + values_fa , _contextManaged = False ) ]
                        lines1 += [ HLine(_contextManaged = False) ]
                    lines1 += [ HLine(_contextManaged = False) ]
                    # Prepare integrated table
                    lines2 = []
                    lines2 += [ HLine(_contextManaged = False) ]
                    lines2 += [ HLine(_contextManaged = False) ]
                    lines2 += [ TableLine( columns = ['',r'$P_{D}[\%]$',r'$F_{a}[\%]$'], _contextManaged = False ) ]
                    lines2 += [ HLine(_contextManaged = False) ]
                    # Get all et/eta bins
                    current_table = table.loc[table.name==name]
                    # Get reference values
                    passed_fa = current_table.reference_background_passed.sum()
                    total_fa = current_table.reference_background_total.sum()
                    fa = passed_fa/total_fa * 100
                    passed_det = current_table.reference_signal_passed.sum()
                    total_det = current_table.reference_signal_total.sum()
                    det = passed_det/total_det * 100
                    lines2 += [ TableLine( columns = ['Reference','%1.2f (%d/%d)'%(det,passed_det,total_det),
                                          '%1.2f (%d/%d)'%(fa,passed_fa,total_fa)]  , _contextManaged = False ) ]
                    # Get corrected values
                    passed_fa = current_table.background_corrected_passed.sum()
                    total_fa = current_table.background_corrected_total.sum()
                    fa = passed_fa/total_fa * 100
                    passed_det = current_table.signal_corrected_passed.sum()
                    total_det = current_table.signal_corrected_total.sum()
                    det = passed_det/total_det * 100
                    lines2 += [ TableLine( columns = [name.replace('_','\_'),'%1.2f (%d/%d)'%(det,passed_det,total_det),
                                          '%1.2f (%d/%d)'%(fa,passed_fa,total_fa)]  , _contextManaged = False ) ]
                    # Get non-corrected values
                    passed_fa = current_table.background_passed.sum()
                    total_fa = current_table.background_total.sum()
                    fa = passed_fa/total_fa * 100
                    passed_det = current_table.signal_passed.sum()
                    total_det = current_table.signal_total.sum()
                    det = passed_det/total_det * 100
                    lines2 += [ TableLine( columns = [name.replace('_','\_')+' (not corrected)','%1.2f (%d/%d)'%(det,passed_det,total_det),
                                            '%1.2f (%d/%d)'%(fa,passed_fa,total_fa)]  , _contextManaged = False ) ]
                    lines2 += [ HLine(_contextManaged = False) ]
                    lines2 += [ HLine(_contextManaged = False) ]
                    with BeamerSlide( title = "Efficiency Values After Correction"  ):
                        with Table( caption = '$P_{d}$ and $F_{a}$ for all phase space regions.') as _table:
                            with ResizeBox( size = 1 ) as rb:
                                with Tabular( columns = 'l' + 'c' * len(etabins_str) ) as tabular:
                                    tabular = tabular
                                    for line in lines1:
                                        if isinstance(line, TableLine):
                                            tabular += line
                                        else:
                                            TableLine(line, rounding = None)
                        with Table( caption = 'Integrated efficiency comparison.') as _table:
                            with ResizeBox( size = 0.6 ) as rb:
                                with Tabular( columns = 'l' + 'c' * 3 ) as tabular:
                                    tabular = tabular
                                    for line in lines2:
                                        if isinstance(line, TableLine):
                                            tabular += line
                                        else:
                                            TableLine(line, rounding = None)



    #
    # Export all models to keras/onnx using the prometheus format
    #
    def export( self, models, model_output_format , conf_output, reference_name, to_onnx=False,
                max_avgmu = 100, min_avgmu = 0):


        try:
            os.makedirs('models')
        except:
            pass

        model_etmin_vec = []
        model_etmax_vec = []
        model_etamin_vec = []
        model_etamax_vec = []
        model_paths = []
        slopes = []
        offsets = []

        # serialize all models
        for etBinIdx in range( len(self.__etbins)-1 ):
            for etaBinIdx in range( len(self.__etabins)-1 ):

                model = models[etBinIdx][etaBinIdx]

                if model['etBinIdx']!=etBinIdx:
                    MSG_FATAL(self, 'Model etBinIdx (%d) is different than etBinIdx (%d). Abort.', model['etBinIdx'], etBinIdx)

                if model['etaBinIdx']!=etaBinIdx:
                    MSG_FATAL(self, 'Model etaBinIdx (%d) is different than etaBinIdx (%d). Abort.', model['etaBinIdx'], etaBinIdx)

                model['min_avgmu'] = min_avgmu
                model['max_avgmu'] = max_avgmu

                model_etmin_vec.append( float(model['etBin'][0]) )
                model_etmax_vec.append( float(model['etBin'][1]) )
                model_etamin_vec.append( float(model['etaBin'][0]) )
                model_etamax_vec.append( float(model['etaBin'][1]) )
                etBinIdx = model['etBinIdx']
                etaBinIdx = model['etaBinIdx']

                model_name = 'models/'+model_output_format%( etBinIdx, etaBinIdx )
                model_paths.append( model_name+'.onnx' ) #default is onnx since this will be used into the athena base
                model_json = model['model'].to_json()
                with open(model_name+".json", "w") as json_file:
                    json_file.write(model_json)
                    # saving the model weight separately
                    model['model'].save_weights(model_name+".h5")

                if to_onnx:
                    # NOTE: This is a hack since I am not be able to convert to onnx inside this function. I need to
                    # open a new python instance (call by system) to convert my models.
                    command = 'convert2onnx.py -j {FILE}.json -w {FILE}.h5 -o {FILE}.onnx'.format(FILE=model_name)
                    os.system(command)


                slopes.append( model['thresholds'][reference_name]['slope'] )
                offsets.append( model['thresholds'][reference_name]['offset'] )


        def list_to_str( l ):
            s = str()
            for ll in l:
              s+=str(ll)+'; '
            return s[:-2]

        # Write the config file
        file = TEnv( 'ringer' )
        file.SetValue( "__name__", 'should_be_filled' )
        file.SetValue( "__version__", 'should_be_filled' )
        file.SetValue( "__operation__", reference_name )
        file.SetValue( "__signature__", 'should_be_filled' )
        file.SetValue( "Model__size"  , str(len(model_paths)) )
        file.SetValue( "Model__etmin" , list_to_str(model_etmin_vec) )
        file.SetValue( "Model__etmax" , list_to_str(model_etmax_vec) )
        file.SetValue( "Model__etamin", list_to_str(model_etamin_vec) )
        file.SetValue( "Model__etamax", list_to_str(model_etamax_vec) )
        file.SetValue( "Model__path"  , list_to_str( model_paths ) )
        file.SetValue( "Threshold__size"  , str(len(model_paths)) )
        file.SetValue( "Threshold__etmin" , list_to_str(model_etmin_vec) )
        file.SetValue( "Threshold__etmax" , list_to_str(model_etmax_vec) )
        file.SetValue( "Threshold__etamin", list_to_str(model_etamin_vec) )
        file.SetValue( "Threshold__etamax", list_to_str(model_etamax_vec) )
        file.SetValue( "Threshold__slope" , list_to_str(slopes) )
        file.SetValue( "Threshold__offset", list_to_str(offsets) )
        file.SetValue( "Threshold__MaxAverageMu", list_to_str([max_avgmu]*len(model_paths)))
        file.SetValue( "Threshold__MinAverageMu", list_to_str([min_avgmu]*len(model_paths)))

        MSG_INFO( self, "Export all tuning configuration to %s.", conf_output)
        file.WriteFile(conf_output)




    #
    # Find the threshold given a reference value
    #
    def find_threshold(self, th1,effref):
        nbins = th1.GetNbinsX()
        fullArea = th1.Integral(0,nbins+1)
        if fullArea == 0:
            return 0,1
        notDetected = 0.0; i = 0
        while (1. - notDetected > effref):
            cutArea = th1.Integral(0,i)
            i+=1
            prevNotDetected = notDetected
            notDetected = cutArea/fullArea
        eff = 1. - notDetected
        prevEff = 1. -prevNotDetected
        deltaEff = (eff - prevEff)
        threshold = th1.GetBinCenter(i-1)+(effref-prevEff)/deltaEff*(th1.GetBinCenter(i)-th1.GetBinCenter(i-1))
        error = 1./np.sqrt(fullArea)
        return threshold, error

    #
    # Get all points in the 2D histogram given a reference value
    #
    def get_points( self, th2 , effref):
        nbinsy = th2.GetNbinsY()
        x = list(); y = list(); errors = list()
        for by in range(nbinsy):
            xproj = th2.ProjectionX('xproj'+str(time.time()),by+1,by+1)
            discr, error = self.find_threshold(xproj,effref)
            dbin = xproj.FindBin(discr)
            x.append(discr); y.append(th2.GetYaxis().GetBinCenter(by+1))
            errors.append( error )
        return x,y,errors



    #
    # Calculate the linear fit given a 2D histogram and reference value and return the slope and offset
    #
    def fit(self, th2,effref):
        x_points, y_points, error_points = self.get_points(th2, effref )
        import array
        from ROOT import TGraphErrors, TF1
        g = TGraphErrors( len(x_points)
                             , array.array('d',y_points,)
                             , array.array('d',x_points)
                             , array.array('d',[0.]*len(x_points))
                             , array.array('d',error_points) )
        firstBinVal = th2.GetYaxis().GetBinLowEdge(th2.GetYaxis().GetFirst())
        lastBinVal = th2.GetYaxis().GetBinLowEdge(th2.GetYaxis().GetLast()+1)
        f1 = TF1('f1','pol1',firstBinVal, lastBinVal)
        g.Fit(f1,"FRq")
        slope = f1.GetParameter(1)
        offset = f1.GetParameter(0)
        return slope, offset, x_points, y_points, error_points


    #
    # Calculate the numerator and denomitator given the 2D histogram and slope/offset parameters
    #
    def calculate_num_and_den_from_hist(self, th2, slope, offset) :

      nbinsy = th2.GetNbinsY()
      th1_num = th2.ProjectionY(th2.GetName()+'_proj'+str(time.time()),1,1)
      th1_num.Reset("ICESM")
      numerator=0; denominator=0
      # Calculate how many events passed by the threshold
      for by in range(nbinsy) :
          xproj = th2.ProjectionX('xproj'+str(time.time()),by+1,by+1)
          # Apply the correction using ax+b formula
          threshold = slope*th2.GetYaxis().GetBinCenter(by+1)+ offset
          dbin = xproj.FindBin(threshold)
          num = xproj.Integral(dbin+1,xproj.GetNbinsX()+1)
          th1_num.SetBinContent(by+1,num)
          numerator+=num
          denominator+=xproj.Integral(-1, xproj.GetNbinsX()+1)

      return numerator/denominator, numerator, denominator


    def calculate_num_and_den_from_output(self, output, avgmu, slope, offset) :
      thr = avgmu*slope + offset
      denominator = len(output)
      numerator = sum( output>thr  )
      return denominator/numerator, numerator, denominator



    #
    # Plot 2D histogram function based on ROOT used by fit_table class
    #
    def plot_2d_hist( self, th2, slope, offset, x_points, y_points, error_points, outname, xlabel='',
                      etBinIdx=None, etaBinIdx=None, etBins=None,etaBins=None, label='Internal',
                      ref_value=None, pd_value=None):



        def toStrBin(etlist = None, etalist = None, etidx = None, etaidx = None):
            if etlist and etidx is not None:
                etlist=copy(etlist)
                if etlist[-1]>9999:  etlist[-1]='#infty'
                binEt = (str(etlist[etidx]) + ' < E_{T} [GeV] < ' + str(etlist[etidx+1]) if etidx+1 < len(etlist) else
                                         'E_{T} > ' + str(etlist[etidx]) + ' GeV')
                return binEt
            if etalist and etaidx is not None:
                binEta = (str(etalist[etaidx]) + ' < |#eta| < ' + str(etalist[etaidx+1]) if etaidx+1 < len(etalist) else
                                            str(etalist[etaidx]) + ' <|#eta| < 2.47')
                return binEta

        canvas = TCanvas("canvas","canvas",500,500)
        rpl.set_figure(canvas)
        gStyle.SetPalette(self.__palette)
        canvas.SetRightMargin(0.15)
        canvas.SetTopMargin(0.15)
        canvas.SetLogz()

        th2.GetXaxis().SetTitle('Neural Network output (Discriminant)')
        th2.GetYaxis().SetTitle(xlabel)
        th2.GetZaxis().SetTitle('Count')
        th2.Draw('colz')

        g = TGraphErrors(len(x_points), array.array('d',x_points), array.array('d',y_points), array.array('d',error_points), array.array('d',[0]*len(x_points)))
        g.SetMarkerColor(kBlue)
        g.SetMarkerStyle(8)
        g.SetMarkerSize(1)
        g.Draw("P same")

        line = TLine(slope*th2.GetYaxis().GetXmin()+offset,th2.GetYaxis().GetXmin(), slope*th2.GetYaxis().GetXmax()+offset, th2.GetYaxis().GetXmax())
        line.SetLineColor(kBlack)
        line.SetLineWidth(2)
        line.Draw()

        # Add text labels into the canvas
        text = toStrBin(etlist=etBins, etidx=etBinIdx)
        text+= ', '+toStrBin(etalist=etaBins, etaidx=etaBinIdx)

        if ref_value and pd_value:
            text+=', P_{D} = %1.2f (%1.2f) [%%]'%(pd_value*100, ref_value*100)

        rpl.add_text(0.15, 0.885, text, textsize=.03)
        rpl.set_atlas_label(0.15, 0.94, label)
        rpl.format_canvas_axes(XLabelSize=16, YLabelSize=16, XTitleOffset=0.87, ZLabelSize=16,ZTitleSize=16, YTitleOffset=0.87, ZTitleOffset=1.1)
        canvas.SaveAs(outname)
        return outname


    @classmethod
    def create_best_sorts_config_table( cls, best_sort_models ):

        d = collections.OrderedDict({
             'et_bin':[], 
             'eta_bin':[], 
             'sort':[], 
             'init':[], 
             'file_name':[],
             'min_avgmu':[],
             'max_avgmu':[], 
             'model_idx':[],
            })

        op_names = best_sort_models[0][0]['thresholds'].keys()
        for op in op_names:
            d[op+'_offset'] = []
            d[op+'_slope']  = []
    
        for et_bin in range(len(best_sort_models)):        
            for eta_bin in range(len(best_sort_models[0])):
            
                model = best_sort_models[et_bin][eta_bin]
                d['et_bin'].append( et_bin )
                d['eta_bin'].append( eta_bin )
                d['sort'].append( model['sort'] )
                d['init'].append( model['init'] )
                d['file_name'].append( model['file_name'] )
                d['model_idx'].append( model['model_idx'] )
                d['min_avgmu'].append( model['min_avgmu'] )
                d['max_avgmu'].append( model['max_avgmu'] )   
                for op in op_names:
                    d[op+'_slope'].append(model['thresholds'][op]['slope'])
                    d[op+'_offset'].append(model['thresholds'][op]['offset'])

        return pd.DataFrame(d)


# Imports
from setuptools import setup, find_packages

# Loading README file
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='kolmov',
    version='2.0.5',
    license='GPL-3.0',
    description='A Framework for performing cross validation for Ringer tunings',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    author='Micael Veríssimo de Araújo, Gabriel Gazola Milan, João Victor da Fonseca Pinto',
    author_email='micael.verissimo@lps.ufrj.br, gabriel.milan@lps.ufrj.br, jodafons@lps.ufrj.br',
    url='https://github.com/micaelverissimo/kolmov',
    keywords=['framework', 'validation', 'machine-learning',
              'ai', 'plotting', 'data-visualization'],
    install_requires=[
        "numpy>=1.16.6,<2.0a0",
        "pandas==1.3.0",
        "matplotlib==3.4.2",
        "seaborn==0.11.1",
        "Gaugi>=1.0.0",
        "tensorflow==2.3.1",
        "scikit-learn==0.24.2",
        "onnxruntime==1.8.0",
        "keras2onnx==1.7.0"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



__all__ = []

# Check if root is installed
has_root=False
try:
    import ROOT
    has_root=True
except:
    print('ROOT not installed at your system. correction table class not available.')


from . import utils
__all__.extend( utils.__all__ )
from .utils import *

from . import crossval_table
__all__.extend( crossval_table.__all__ )
from .crossval_table import *

from . import fit_table
__all__.extend( fit_table.__all__ )
from .fit_table import *

from . import crossval_fit_table
__all__.extend( crossval_fit_table.__all__ )
from .crossval_fit_table import *





__all__ = []


from . import constants
__all__.extend( constants.__all__ )
from .constants import *

from . import utils
__all__.extend( utils.__all__ )
from .utils import *

from . import plot_functions
__all__.extend( plot_functions.__all__ )
from .plot_functions import *

#from . import legacy_exports
#__all__.extend( legacy_exports.__all__ )
#from .legacy_exports import *


